import os
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, callbacks
from keras_cv_attention_models.coco import anchors_func
from tqdm import tqdm


@backend.register_keras_serializable(package="kecam/coco")
class DecodePredictions(layers.Layer):
    """
    The most simple version decoding prediction and NMS:

    >>> from keras_cv_attention_models import efficientdet, test_images
    >>> model = efficientdet.EfficientDetD0()
    >>> preds = model(model.preprocess_input(test_images.dog()))

    # Decode and NMS
    >>> from keras_cv_attention_models import coco
    >>> input_shape = model.input_shape[1:-1]
    >>> anchors = coco.get_anchors(input_shape=input_shape, pyramid_levels=[3, 7], anchor_scale=4)
    >>> dd = coco.decode_bboxes(preds[0], anchors).numpy()
    >>> rr = tf.image.non_max_suppression(dd[:, :4], dd[:, 4:].max(-1), score_threshold=0.3, max_output_size=15, iou_threshold=0.5)
    >>> dd_nms = tf.gather(dd, rr).numpy()
    >>> bboxes, labels, scores = dd_nms[:, :4], dd_nms[:, 4:].argmax(-1), dd_nms[:, 4:].max(-1)
    >>> print(f"{bboxes = }, {labels = }, {scores = }")
    >>> # bboxes = array([[0.433231  , 0.54432285, 0.8778939 , 0.8187578 ]], dtype=float32), labels = array([17]), scores = array([0.85373735], dtype=float32)
    """

    def __init__(
        self,
        input_shape=512,
        pyramid_levels=[3, 7],
        anchors_mode=None,
        use_object_scores="auto",
        anchor_scale="auto",
        aspect_ratios=(1, 2, 0.5),
        num_scales=3,
        score_threshold=0.3,  # decode parameter, can be set new value in `self.call`
        iou_or_sigma=0.5,  # decode parameter, can be set new value in `self.call`
        max_output_size=100,  # decode parameter, can be set new value in `self.call`
        method="hard",  # decode parameter, can be set new value in `self.call`
        mode="global",  # decode parameter, can be set new value in `self.call`
        topk=0,  # decode parameter, can be set new value in `self.call`
        use_static_output=False,  # Set to True if using this as an actual layer, especially for converting tflite
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.pyramid_levels = list(range(min(pyramid_levels), max(pyramid_levels) + 1))
        use_object_scores, num_anchors, anchor_scale = anchors_func.get_anchors_mode_parameters(anchors_mode, use_object_scores, "auto", anchor_scale)
        self.aspect_ratios, self.num_scales = aspect_ratios, num_scales
        self.anchors_mode, self.use_object_scores, self.anchor_scale = anchors_mode, use_object_scores, anchor_scale  # num_anchors not using
        if input_shape is not None and (isinstance(input_shape, (list, tuple)) and input_shape[1] is not None):
            self.__init_anchor__(input_shape)
        else:
            self.anchors = None
        self.__input_shape__ = input_shape
        self.use_static_output = use_static_output
        self.nms_kwargs = {
            "score_threshold": score_threshold,
            "iou_or_sigma": iou_or_sigma,
            "max_output_size": max_output_size,
            "method": method,
            "mode": mode,
            "topk": topk,
        }
        super().build(input_shape)

    def __init_anchor__(self, input_shape):
        if isinstance(input_shape, (list, tuple)):
            # input_shape = input_shape[:2] if backend.image_data_format() == "channels_last" else input_shape[-2:]
            channel_axis, channel_dim = min(enumerate(input_shape), key=lambda xx: xx[1])  # Assume the smallest value is the channel dimension
            input_shape = [dim for axis, dim in enumerate(input_shape) if axis != channel_axis]
        else:
            input_shape = (input_shape, input_shape)

        if self.anchors_mode == anchors_func.ANCHOR_FREE_MODE:
            self.anchors = anchors_func.get_anchor_free_anchors(input_shape, self.pyramid_levels)
        elif self.anchors_mode == anchors_func.YOLOR_MODE:
            self.anchors = anchors_func.get_yolor_anchors(input_shape, self.pyramid_levels)
        else:
            grid_zero_start = False
            self.anchors = anchors_func.get_anchors(input_shape, self.pyramid_levels, self.aspect_ratios, self.num_scales, self.anchor_scale, grid_zero_start)
        self.__input_shape__ = input_shape
        return self.anchors

    def __topk_class_boxes_single__(self, pred, topk=5000):
        # https://github.com/google/automl/tree/master/efficientdet/tf2/postprocess.py#L82
        bbox_outputs, class_outputs = pred[:, :4], pred[:, 4:]
        num_classes = class_outputs.shape[-1]
        class_outputs_flatten = functional.reshape(class_outputs, -1)
        topk = class_outputs_flatten.shape[0] if topk == -1 else topk  # select all if -1
        _, class_topk_indices = functional.top_k(class_outputs_flatten, k=topk, sorted=False)
        # get original indices for class_outputs, original_indices_hh -> picking indices, original_indices_ww -> picked labels
        original_indices_hh, original_indices_ww = class_topk_indices // num_classes, class_topk_indices % num_classes
        class_indices = functional.stack([original_indices_hh, original_indices_ww], axis=-1)
        scores_topk = functional.gather_nd(class_outputs, class_indices)
        bboxes_topk = functional.gather(bbox_outputs, original_indices_hh)
        return bboxes_topk, scores_topk, original_indices_ww, original_indices_hh

    # def __nms_per_class__(self, bbs, ccs, labels, score_threshold=0.3, iou_threshold=0.5, soft_nms_sigma=0.5, max_output_size=100):
    #     # https://github.com/google/automl/tree/master/efficientdet/tf2/postprocess.py#L409
    #     # Not using, same result with `torchvision.ops.batched_nms`
    #     rrs = []
    #     for ii in tf.unique(labels)[0]:
    #         pick = tf.where(labels == ii)
    #         bb, cc = tf.gather_nd(bbs, pick), tf.gather_nd(ccs, pick)
    #         rr, nms_scores = tf.image.non_max_suppression_with_scores(bb, cc, max_output_size, iou_threshold, score_threshold, soft_nms_sigma)
    #         bb_nms = tf.gather(bb, rr)
    #         rrs.append(tf.concat([bb_nms, tf.ones([bb_nms.shape[0], 1]) * tf.cast(ii, bb_nms.dtype), tf.expand_dims(nms_scores, 1)], axis=-1))
    #     rrs = tf.concat(rrs, axis=0)
    #     if tf.shape(rrs)[0] > max_output_size:
    #         score_top_k = tf.argsort(rrs[:, -1], direction="DESCENDING")[:max_output_size]
    #         rrs = tf.gather(rrs, score_top_k)
    #     bboxes, labels, scores = rrs[:, :4], rrs[:, 4], rrs[:, -1]
    #     return bboxes.numpy(), labels.numpy(), scores.numpy()

    def __nms_per_class__(self, bbs, ccs, labels, score_threshold=0.3, iou_threshold=0.5, soft_nms_sigma=0.5, max_output_size=100):
        # From torchvision.ops.batched_nms strategy: in order to perform NMS independently per class. we add an offset to all the boxes.
        # The offset is dependent only on the class idx, and is large enough so that boxes from different classes do not overlap
        # Same result with per_class method: https://github.com/google/automl/tree/master/efficientdet/tf2/postprocess.py#L409
        cls_offset = functional.cast(labels, bbs.dtype) * (functional.reduce_max(bbs) + 1)
        bbs_per_class = bbs + functional.expand_dims(cls_offset, -1)
        rr, nms_scores = functional.non_max_suppression_with_scores(bbs_per_class, ccs, max_output_size, iou_threshold, score_threshold, soft_nms_sigma)
        return functional.gather(bbs, rr), functional.gather(labels, rr), nms_scores

    def __nms_global__(self, bbs, ccs, labels, score_threshold=0.3, iou_threshold=0.5, soft_nms_sigma=0.5, max_output_size=100):
        rr, nms_scores = functional.non_max_suppression_with_scores(bbs, ccs, max_output_size, iou_threshold, score_threshold, soft_nms_sigma)
        return functional.gather(bbs, rr), functional.gather(labels, rr), nms_scores

    def __object_score_split__(self, pred):
        return pred[:, :-1], pred[:, -1]  # May overwrite

    def __to_static__(self, bboxs, lables, confidences, max_output_size=100):
        indices = functional.expand_dims(functional.range(functional.shape(bboxs)[0]), -1)
        lables = functional.cast(lables, bboxs.dtype)
        concated = functional.concat([bboxs, functional.expand_dims(lables, -1), functional.expand_dims(confidences, -1)], axis=-1)
        concated = functional.tensor_scatter_nd_update(functional.zeros([max_output_size, concated.shape[-1]], dtype=bboxs.dtype), indices, concated)
        return concated

    def __decode_single__(self, pred, score_threshold=0.3, iou_or_sigma=0.5, max_output_size=100, method="hard", mode="global", topk=0, input_shape=None):
        # https://github.com/google/automl/tree/master/efficientdet/tf2/postprocess.py#L159
        pred = functional.cast(pred.detach() if hasattr(pred, "detach") else pred, "float32")
        if input_shape is not None:
            self.__init_anchor__(input_shape)

        if self.use_object_scores:  # YOLO outputs: [bboxes, classses_score, object_score]
            pred, object_scores = self.__object_score_split__(pred)

        if topk != 0:
            bbs, ccs, labels, picking_indices = self.__topk_class_boxes_single__(pred, topk)
            anchors = functional.gather(self.anchors, picking_indices)
            if self.use_object_scores:
                ccs = ccs * functional.gather(object_scores, picking_indices)
        else:
            bbs, ccs, labels = pred[:, :4], functional.reduce_max(pred[:, 4:], axis=-1), functional.argmax(pred[:, 4:], axis=-1)
            anchors = self.anchors
            if self.use_object_scores:
                ccs = ccs * object_scores

        # print(f"{bbs.shape = }, {anchors.shape = }")
        bbs_decoded = anchors_func.decode_bboxes(bbs, anchors)
        iou_threshold, soft_nms_sigma = (1.0, iou_or_sigma / 2) if method.lower() == "gaussian" else (iou_or_sigma, 0.0)

        if mode == "per_class":
            bboxs, lables, confidences = self.__nms_per_class__(bbs_decoded, ccs, labels, score_threshold, iou_threshold, soft_nms_sigma, max_output_size)
        else:
            bboxs, lables, confidences = self.__nms_global__(bbs_decoded, ccs, labels, score_threshold, iou_threshold, soft_nms_sigma, max_output_size)

        return self.__to_static__(bboxs, lables, confidences, max_output_size) if self.use_static_output else (bboxs, lables, confidences)

    def call(self, preds, input_shape=None, training=False, **nms_kwargs):
        """
        https://github.com/google/automl/tree/master/efficientdet/tf2/postprocess.py#L159

        input_shape: actual input shape if model using dynamic input shape `[None, None, 3]`.
        nms_kwargs:
          score_threshold: float value in (0, 1), min score threshold, lower output score will be excluded. Default 0.3.
          iou_or_sigma: means `soft_nms_sigma` if method is "gaussian", else `iou_threshold`. Default 0.5.
          max_output_size: max output size for `tf.image.non_max_suppression_with_scores`. Default 100.
              If use_static_output=True, fixed output shape will be `[batch, max_output_size, 6]`.
          method: "gaussian" or "hard".  Default "hard".
          mode: "global" or "per_class". "per_class" is strategy from `torchvision.ops.batched_nms`. Default "global".
          topk: Using topk highest scores, each bbox may have multi labels. Set `0` to disable, `-1` using all. Default 0.
        """
        self.nms_kwargs.update(nms_kwargs)
        if self.use_static_output:
            return functional.map_fn(lambda xx: self.__decode_single__(xx, **nms_kwargs), preds)
        elif len(preds.shape) == 3:
            return [self.__decode_single__(pred, **self.nms_kwargs, input_shape=input_shape) for pred in preds]
        else:
            return self.__decode_single__(preds, **self.nms_kwargs, input_shape=input_shape)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "input_shape": self.__input_shape__,
                "pyramid_levels": self.pyramid_levels,
                "anchors_mode": self.anchors_mode,
                "use_object_scores": self.use_object_scores,
                "anchor_scale": self.anchor_scale,
                "aspect_ratios": self.aspect_ratios,
                "num_scales": self.num_scales,
                "use_static_output": self.use_static_output,
            }
        )
        config.update(self.nms_kwargs)
        return config


def scale_bboxes_back_single(bboxes, image_shape, scale, pad_top, pad_left, target_shape):
    # height, width = target_shape[0] / scale, target_shape[1] / scale
    # bboxes *= [height, width, height, width]
    bboxes *= [target_shape[0], target_shape[1], target_shape[0], target_shape[1]]
    bboxes -= [pad_top, pad_left, pad_top, pad_left]
    bboxes /= scale
    bboxes = functional.clip_by_value(bboxes, 0, clip_value_max=[image_shape[0], image_shape[1], image_shape[0], image_shape[1]])
    # [top, left, bottom, right] -> [left, top, width, height]
    bboxes = functional.stack([bboxes[:, 1], bboxes[:, 0], bboxes[:, 3] - bboxes[:, 1], bboxes[:, 2] - bboxes[:, 0]], axis=-1)
    return bboxes


def image_process(image, target_shape, mean, std, resize_method="bilinear", resize_antialias=False, use_bgr_input=False, letterbox_pad=-1):
    from keras_cv_attention_models.coco import data

    if len(image.shape) < 2:
        image = data.tf_imread(image)  # it's image path
    original_image_shape = functional.shape(image)[:2]
    image = functional.cast(image, "float32")
    image = (image - mean) / std  # automl behavior: rescale -> resize
    image, scale, pad_top, pad_left = data.aspect_aware_resize_and_crop_image(
        image, target_shape, letterbox_pad=letterbox_pad, method=resize_method, antialias=resize_antialias
    )
    if use_bgr_input:
        image = image[:, :, ::-1]
    return image, scale, pad_top, pad_left, original_image_shape


def init_eval_dataset(
    data_name="coco/2017",
    input_shape=(512, 512),
    batch_size=8,
    rescale_mode="torch",
    resize_method="bilinear",
    resize_antialias=False,
    letterbox_pad=-1,
    use_bgr_input=False,
):
    import tensorflow_datasets as tfds
    from keras_cv_attention_models.coco import data

    dataset = data.detection_dataset_from_custom_json(data_name) if data_name.endswith(".json") else tfds.load(data_name)
    ds = dataset.get("validation", dataset.get("test", None))

    mean, std = data.init_mean_std_by_rescale_mode(rescale_mode)
    __image_process__ = lambda image: image_process(image, input_shape, mean, std, resize_method, resize_antialias, use_bgr_input, letterbox_pad)
    # ds: [resized_image, scale, pad_top, pad_left, original_image_shape, image_id]
    ds = ds.map(lambda datapoint: (*__image_process__(datapoint["image"]), datapoint.get("image/id", datapoint["image"])))
    ds = ds.batch(batch_size)
    return ds


def model_detection_and_decode(model, eval_dataset, pred_decoder, nms_kwargs={}, is_coco=True, image_id_map=None):
    from keras_cv_attention_models.coco import data

    target_shape = (eval_dataset.element_spec[0].shape[1], eval_dataset.element_spec[0].shape[2])
    num_classes = model.output_shape[-1] - 4
    if is_coco:
        to_91_labels = (lambda label: label + 1) if num_classes >= 90 else (lambda label: data.COCO_80_to_90_LABEL_DICT[label] + 1)
    else:
        to_91_labels = lambda label: label
    # Format: [image_id, x, y, width, height, score, class]
    to_coco_eval_single = lambda image_id, bbox, label, score: [image_id, *bbox.tolist(), score, to_91_labels(label)]

    results = []
    for images, scales, pad_tops, pad_lefts, original_image_shapes, image_ids in tqdm(eval_dataset):
        preds = model(images)
        preds = functional.cast(preds, "float32")
        # decoded_preds: [[bboxes, labels, scores], [bboxes, labels, scores], ...]
        decoded_preds = pred_decoder(preds, **nms_kwargs)

        # Loop on batch
        for rr, image_shape, scale, pad_top, pad_left, image_id in zip(decoded_preds, original_image_shapes, scales, pad_tops, pad_lefts, image_ids):
            bboxes, labels, scores = rr
            image_id, bboxes, labels, scores = image_id.numpy(), bboxes.numpy(), labels.numpy(), scores.numpy()
            if image_id_map is not None:
                image_id = image_id_map[image_id.decode() if isinstance(image_id, bytes) else image_id]
            bboxes = scale_bboxes_back_single(bboxes, image_shape, scale, pad_top, pad_left, target_shape).numpy()
            results.extend([to_coco_eval_single(image_id, bb, cc, ss) for bb, cc, ss in zip(bboxes, labels, scores)])  # Loop on prediction results
    return functional.convert_to_tensor(results).numpy()


class COCOEvaluation:
    def __init__(self, annotations=None):
        from pycocotools.coco import COCO

        if annotations is None:
            url = "https://github.com/leondgarse/keras_cv_attention_models/releases/download/efficientdet/coco_annotations_instances_val2017.json"
            annotations = backend.get_file(origin=url)

        if isinstance(annotations, dict):  # json already loaded as dict
            coco_gt = COCO()
            coco_gt.dataset = annotations
            coco_gt.createIndex()
        else:
            coco_gt = COCO(annotations)
        self.coco_gt = coco_gt

    def __call__(self, detection_results):
        from pycocotools.cocoeval import COCOeval

        image_ids = [ii["image_id"] for ii in detection_results] if isinstance(detection_results[0], dict) else [ii[0] for ii in detection_results]
        image_ids = list(set(image_ids))
        print("len(image_ids) =", len(image_ids))
        coco_dt = self.coco_gt.loadRes(detection_results)
        coco_eval = COCOeval(cocoGt=self.coco_gt, cocoDt=coco_dt, iouType="bbox")
        coco_eval.params.imgIds = image_ids
        coco_eval.evaluate()
        coco_eval.accumulate()
        coco_eval.summarize()
        return coco_eval


def to_coco_json(detection_results, save_path, indent=2):
    import json

    __to_coco_json__ = lambda xx: {"image_id": int(xx[0]), "bbox": [float(ii) for ii in xx[1:5]], "score": float(xx[5]), "category_id": int(xx[6])}
    aa = [__to_coco_json__(ii) for ii in detection_results]
    with open(save_path, "w") as ff:
        json.dump(aa, ff, indent=indent)


def to_coco_annotation(json_path):
    import json
    from PIL import Image

    with open(json_path, "r") as ff:
        aa = json.load(ff)

    # int conversion just in case key is str
    categories = {int(kk): vv for kk, vv in aa["indices_2_labels"].items()} if "indices_2_labels" in aa else {}
    annotations, images, image_id_map = [], [], {}
    for image_id, ii in enumerate(aa.get("validation", aa.get("test", []))):
        width, height = Image.open(ii["image"]).size  # For decoding bboxes, not actually openning images
        for bb, label in zip(ii["objects"]["bbox"], ii["objects"]["label"]):
            # bb [top, left, bottom, right] in [0, 1] -> [left, top, bbox_width, bbox_height] with actual coordinates
            top = bb[0] * height
            left = bb[1] * width
            bbox_height = bb[2] * height - top
            bbox_width = bb[3] * width - left
            bb = [left, top, bbox_width, bbox_height]
            area = bbox_width * bbox_height  # Actual area in COCO is the segmentation area, doesn't matter in detection mission

            label = int(label)
            annotations.append({"bbox": bb, "category_id": label, "image_id": image_id, "id": len(annotations), "iscrowd": 0, "area": area})
            if label not in categories:
                categories[label] = str(len(categories))
        images.append({"id": image_id, "file_name": ii["image"], "height": height, "width": width})
        image_id_map[ii["image"]] = image_id
    categories = [{"id": kk, "name": vv} for kk, vv in categories.items()]
    return {"images": images, "annotations": annotations, "categories": categories}, image_id_map


# Wrapper a callback for using in training
class COCOEvalCallback(callbacks.Callback):
    """
    Basic test:
    >>> from keras_cv_attention_models import efficientdet, coco
    >>> model = efficientdet.EfficientDetD0()
    >>> ee = coco.eval_func.COCOEvalCallback(batch_size=4, model_basic_save_name='test', rescale_mode='raw', anchors_mode="anchor_free")
    >>> ee.model = model
    >>> ee.on_epoch_end()
    """

    def __init__(
        self,
        data_name="coco/2017",  # [init_eval_dataset parameters]
        batch_size=8,
        resize_method="bilinear",
        resize_antialias=False,
        rescale_mode="auto",
        letterbox_pad=-1,
        use_bgr_input=False,
        take_samples=-1,
        nms_score_threshold=0.001,  # [model_detection_and_decode parameters]
        nms_iou_or_sigma=0.5,
        nms_max_output_size=100,
        nms_method="gaussian",
        nms_mode="per_class",
        nms_topk=5000,
        anchors_mode="auto",  # [model anchors related parameters]
        anchor_scale=4,  # Init anchors for model prediction. "auto" means 1 if (anchors_mode=="anchor_free" or anchors_mode=="yolor"), else 4
        aspect_ratios=(1, 2, 0.5),  # For efficientdet anchors only
        num_scales=3,  # For efficientdet anchors only
        annotation_file=None,
        save_json=None,
        start_epoch=0,  # [trainign callbacks parameters]
        frequency=1,
        model_basic_save_name=None,
    ):
        super().__init__()
        self.anchors_mode = anchors_mode
        self.take_samples, self.annotation_file, self.start_epoch, self.frequency = take_samples, annotation_file, start_epoch, frequency
        self.save_json, self.model_basic_save_name, self.save_path, self.item_key = save_json, model_basic_save_name, "checkpoints", "val_ap_ar"
        self.data_name = data_name

        self.dataset_kwargs = {
            "data_name": data_name,
            "batch_size": batch_size,
            "rescale_mode": rescale_mode,
            "resize_method": resize_method,
            "resize_antialias": resize_antialias,
            "letterbox_pad": letterbox_pad,
            "use_bgr_input": use_bgr_input,
        }
        self.nms_kwargs = {
            "score_threshold": nms_score_threshold,
            "iou_or_sigma": nms_iou_or_sigma,
            "max_output_size": nms_max_output_size,
            "method": nms_method,
            "mode": nms_mode,
            "topk": nms_topk,
        }
        self.anchor_kwargs = {
            "anchor_scale": anchor_scale,
            "aspect_ratios": aspect_ratios,
            "num_scales": num_scales,
        }
        self.efficient_det_num_anchors = len(aspect_ratios) * num_scales

        self.is_coco = True if data_name.startswith("coco") and not data_name.endswith(".json") else False
        if self.data_name.endswith(".json") and self.annotation_file is None:
            self.annotation_file, self.image_id_map = to_coco_annotation(self.data_name)
        else:
            self.image_id_map = None

        self.built = False

    def build(self, input_shape, output_shape):
        input_shape = (int(input_shape[1]), int(input_shape[2]))
        self.eval_dataset = init_eval_dataset(input_shape=input_shape, **self.dataset_kwargs)
        if self.anchors_mode is None or self.anchors_mode == "auto":
            self.anchors_mode, num_anchors = anchors_func.get_anchors_mode_by_anchors(input_shape, total_anchors=output_shape[1])
        elif self.anchors_mode == anchors_func.EFFICIENTDET_MODE:
            num_anchors = self.efficient_det_num_anchors
        else:
            num_anchors = anchors_func.NUM_ANCHORS.get(self.anchors_mode, 9)
        pyramid_levels = anchors_func.get_pyramid_levels_by_anchors(input_shape, total_anchors=output_shape[1], num_anchors=num_anchors)
        print("\n>>>> [COCOEvalCallback] input_shape: {}, pyramid_levels: {}, anchors_mode: {}".format(input_shape, pyramid_levels, self.anchors_mode))
        # print(">>>>", self.dataset_kwargs)
        # print(">>>>", self.nms_kwargs)

        self.pred_decoder = DecodePredictions(input_shape, pyramid_levels, self.anchors_mode, **self.anchor_kwargs)

        # Training saving best
        if self.model_basic_save_name is not None:
            self.monitor_save = os.path.join(self.save_path, self.model_basic_save_name + "_epoch_{}_" + self.item_key + "_{}.h5")
            self.monitor_save_re = self.monitor_save.format("*", "*")
            self.is_better = lambda cur, pre: cur >= pre
            self.pre_best = -1e5

        self.coco_evaluation = COCOEvaluation(self.annotation_file)
        self.built = True

    def on_epoch_end(self, epoch=0, logs=None):
        if backend.is_tensorflow_backend:
            from tensorflow.io.gfile import glob
        else:
            from glob2 import glob

        if not self.built:
            if self.dataset_kwargs["rescale_mode"] == "auto":
                self.dataset_kwargs["rescale_mode"] = getattr(self.model, "rescale_mode", "torch")
            self.build(self.model.input_shape, self.model.output_shape)

        if epoch + 1 < self.start_epoch or epoch % self.frequency != 0:
            return

        # pred_decoder = self.model.decode_predictions
        eval_dataset = self.eval_dataset.take(self.take_samples) if self.take_samples > 0 else self.eval_dataset
        detection_results = model_detection_and_decode(self.model, eval_dataset, self.pred_decoder, self.nms_kwargs, self.is_coco, self.image_id_map)
        try:
            coco_eval = self.coco_evaluation(detection_results)
        except:
            print(">>>> Error in running coco_evaluation")
            coco_eval = None
            data_name = self.data_name.replace("/", "_")
            self.save_json = "{}_{}_detection_results_error.json".format(self.model.name, data_name) if self.save_json is None else self.save_json

        if self.save_json is not None:
            to_coco_json(detection_results, self.save_json)
            print(">>>> Detection results saved to:", self.save_json)

        if hasattr(self.model, "history") and hasattr(self.model.history, "history"):
            self.model.history.history.setdefault(self.item_key, []).append(coco_eval.stats.tolist())

        # Training save best
        cur_ap = coco_eval.stats[0] if coco_eval is not None else 0
        if self.model_basic_save_name is not None and self.is_better(cur_ap, self.pre_best):
            self.pre_best = cur_ap
            pre_monitor_saves = glob(self.monitor_save_re)
            # tf.print(">>>> pre_monitor_saves:", pre_monitor_saves)
            if len(pre_monitor_saves) != 0:
                os.remove(pre_monitor_saves[0])
            monitor_save = self.monitor_save.format(epoch + 1, "{:.4f}".format(cur_ap))
            print("\n>>>> Save best to:", monitor_save)
            if self.model is not None:
                self.model.save(monitor_save)

        return coco_eval



from keras_cv_attention_models.resnest.resnest import ResNest, ResNest50, ResNest101, ResNest200, ResNest269, rsoftmax, split_attention_conv2d


__head_doc__ = """
Keras implementation of [ResNeSt](https://github.com/zhanghang1989/ResNeSt).
Paper [PDF 2004.08955 ResNeSt: Split-Attention Networks](https://arxiv.org/pdf/2004.08955.pdf).
"""

__tail_doc__ = """  groups: controls number of split groups in `split_attention_conv2d`.
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
      Set `(None, None, 3)` for dynamic input resolution.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `relu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `softmax`.
  pretrained: one of `None` (random initialization) or 'imagenet' (pre-training on ImageNet).
      Will try to download and load pre-trained model weights if not None.
  **kwargs: other parameters from `AotNet` if not conflict.

Returns:
    A `keras.Model` instance.
"""

ResNest.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model          | Params | FLOPs  | Input | Top1 Acc |
  | -------------- | ------ | ------ | ----- | -------- |
  | resnest50      | 28M    | 5.38G  | 224   | 81.03    |
  | resnest101     | 49M    | 13.33G | 256   | 82.83    |
  | resnest200     | 71M    | 35.55G | 320   | 83.84    |
  | resnest269     | 111M   | 77.42G | 416   | 84.54    |
"""

ResNest50.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

ResNest101.__doc__ = ResNest50.__doc__
ResNest200.__doc__ = ResNest50.__doc__
ResNest269.__doc__ = ResNest50.__doc__

split_attention_conv2d.__doc__ = __head_doc__ + """
Split-Attention. Callable function, NOT defined as a layer.
Generating `attention_scores` using grouped `Conv2D`.

Args:
  inputs: input tensor.
  filters: output dimension.
  kernel_size: kernel size for grouped Conv2D.
  strides: strides for grouped Conv2D.
  groups: number of splitted groups.
  activation: activation used after `BatchNormalization`.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([28, 28, 192])
>>> nn = attention_layers.split_attention_conv2d(inputs, 384)
>>> dd = keras.models.Model(inputs, nn)
>>> dd.summary()
>>> dd.output_shape
(None, 28, 28, 384)

>>> {ii.name: ii.shape for ii in dd.weights}
{'1_g1_conv/kernel:0': TensorShape([3, 3, 96, 384]),
 '1_g2_conv/kernel:0': TensorShape([3, 3, 96, 384]),
 '1_bn/gamma:0': TensorShape([768]),
 '1_bn/beta:0': TensorShape([768]),
 '1_bn/moving_mean:0': TensorShape([768]),
 '1_bn/moving_variance:0': TensorShape([768]),
 '2_conv/kernel:0': TensorShape([1, 1, 384, 96]),
 '2_conv/bias:0': TensorShape([96]),
 '2_bn/gamma:0': TensorShape([96]),
 '2_bn/beta:0': TensorShape([96]),
 '2_bn/moving_mean:0': TensorShape([96]),
 '2_bn/moving_variance:0': TensorShape([96]),
 '3_conv/kernel:0': TensorShape([1, 1, 96, 768]),
 '3_conv/bias:0': TensorShape([768])}
"""

rsoftmax.__doc__ = __head_doc__ + """
Perform group split softmax

input: `[batch, 1, 1, channel]`.
output: `[batch, 1, 1, channel]`.

Args:
  inputs: Input tensor.
  groups: groups to split on channel dimension.
"""



from keras_cv_attention_models.moganet.moganet import MogaNet, MogaNetXtiny, MogaNetTiny, MogaNetSmall, MogaNetBase, MogaNetLarge

__head_doc__ = """
Keras implementation of [Github Westlake-AI/MogaNet](https://github.com/Westlake-AI/MogaNet).
Paper [PDF 2211.03295 Efficient Multi-order Gated Aggregation Network](https://arxiv.org/pdf/2211.03295.pdf).
"""

__tail_doc__ = """  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation for non-attention blocks, default `gelu`.
  attn_activation: activation for attention blocks, default `swish`. `None` for same with `activation`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be a constant value like `0.2`,
      or a tuple value like `(0, 0.2)` indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  layer_scale: int value indicates layer scale init value for each stack. Default `1e-5`, 0 for not using.
      [Going deeper with Image Transformers](https://arxiv.org/abs/2103.17239).
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `None`.
  pretrained: one of `None` (random initialization) or 'imagenet' (pre-training on ImageNet).
      Will try to download and load pre-trained model weights if not None.

Returns:
    A `keras.Model` instance.
"""

MogaNet.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  out_channels: output channels for each stack.
  mlp_ratios: int or list value indicates expand ratio for mlp blocks hidden channel in each stack.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model        | Params | FLOPs  | Input | Top1 Acc |
  | ------------ | ------ | ------ | ----- | -------- |
  | MogaNetXtiny | 2.96M  | 806M   | 224   | 76.5     |
  | MogaNetTiny  | 5.20M  | 1.11G  | 224   | 79.0     |
  |              | 5.20M  | 1.45G  | 256   | 79.6     |
  | MogaNetSmall | 25.3M  | 4.98G  | 224   | 83.4     |
  | MogaNetBase  | 43.7M  | 9.96G  | 224   | 84.2     |
  | MogaNetLarge | 82.5M  | 15.96G | 224   | 84.6     |
"""

MogaNetXtiny.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

MogaNetTiny.__doc__ = MogaNetXtiny.__doc__
MogaNetSmall.__doc__ = MogaNetXtiny.__doc__
MogaNetBase.__doc__ = MogaNetXtiny.__doc__
MogaNetLarge.__doc__ = MogaNetXtiny.__doc__



from keras_cv_attention_models import backend

from keras_cv_attention_models.imagenet.eval_func import (
    evaluation,
    plot_hists,
    combine_hist_into_one,
    parse_timm_log,
    decode_predictions,
    decode_predictions_imagenet21k,
)

if backend.is_tensorflow_backend:
    from keras_cv_attention_models.imagenet.train_func import (
        compile_model,
        init_global_strategy,
        init_loss,
        init_lr_scheduler,
        init_model,
        init_optimizer,
        is_decoupled_weight_decay,
        train,
    )
    from keras_cv_attention_models.imagenet import data
    from keras_cv_attention_models.imagenet.data import init_dataset
    from keras_cv_attention_models.imagenet import callbacks
    from keras_cv_attention_models.imagenet import losses
    from keras_cv_attention_models.imagenet import token_label

    data.random_crop_fraction.__doc__ = """ https://github.com/tensorflow/models/blob/master/official/vision/image_classification/preprocessing.py
    RandomResizedCrop related function.

    For hh_crop = height, max_ww_crop = height * ratio[1], max_area_crop_1 = height * height * ratio[1]
    For ww_crop = width, max_hh_crop = width / ratio[0], max_area_crop_2 = width * width / ratio[0]
    ==> scale_max < min(max_area_crop_1, max_area_crop_2, scale[1])

    As target_area selected:
    For ww_crop = width, max_aspect_ratio = width / (target_area / width) = width * width / target_area
    For hh_crop = height, min_aspect_ratio = (target_area / height) / height = target_area / (height * height)
    ==> ratio in range (min_aspect_ratio, max_aspect_ratio)

    Result:
    ww_crop * hh_crop = target_area
    ww_crop / hh_crop = aspect_ratio
    ==> ww_crop = int(round(math.sqrt(target_area * aspect_ratio)))
        hh_crop = int(round(math.sqrt(target_area / aspect_ratio)))

    As outputs are converted int, for running 1e5 times, results are not exactly in scale and ratio range:
    >>> from keras_cv_attention_models.imagenet import data
    >>> aa = np.array([data.random_crop_fraction(size=(100, 100), ratio=(0.75, 4./3)) for _ in range(100000)])
    >>> hhs, wws = aa[:, 0], aa[:, 1]
    >>> print("Scale range:", ((hhs * wws).min() / 1e4, (hhs * wws).max() / 1e4))
    # Scale range: (0.075, 0.9801)
    >>> print("Ratio range:", ((wws / hhs).min(), (wws / hhs).max()))
    # Ratio range: (0.7272727272727273, 1.375)

    >>> fig, axes = plt.subplots(4, 1, figsize=(6, 8))
    >>> pp = {
    >>>     "ratio distribute": wws / hhs,
    >>>     "scale distribute": wws * hhs / 1e4,
    >>>     "height distribute": hhs,
    >>>     "width distribute": wws,
    >>> }
    >>> for ax, kk in zip(axes, pp.keys()):
    >>>     _ = ax.hist(pp[kk], bins=1000, label=kk)
    >>>     ax.set_title(kk)
    >>> fig.tight_layout()

    Args:
      size (tuple of int): input image shape. `area = size[0] * size[1]`.
      scale (tuple of float): scale range of the cropped image. target_area in range `(scale[0] * area, sacle[1] * area)`.
      ratio (tuple of float): aspect ratio range of the cropped image. cropped `width / height`  in range `(ratio[0], ratio[1])`.

    Returns: cropped size `hh_crop, ww_crop`.
    """

    data.init_mean_std_by_rescale_mode.__doc__ = """
    Args:
      rescale_mode: one of ["tf", "torch", "raw01", "raw"].
        - "tf": mean 127.5, std: 128.0, converts [0, 255] -> [-1, 1].
        - "torch": mean [0.485, 0.456, 0.406] * 255.0, std [0.229, 0.224, 0.225] * 255.0.
        - "raw01": mean 0, std 127.5, converts [0, 255] -> [0, 1].
        - "raw": mean 0, std 1, raw output [0, 255].

    Returns: mean, std
    """

    data.mixup.__doc__ = """ Applies Mixup regularization to a batch of images and labels.
    [1] Hongyi Zhang, Moustapha Cisse, Yann N. Dauphin, David Lopez-Paz
    Mixup: Beyond Empirical Risk Minimization.
    ICLR'18, https://arxiv.org/abs/1710.09412
    """
    data.cutmix.__doc__ = """ Copied and modified from https://keras.io/examples/vision/cutmix/

    Example:
    >>> from keras_cv_attention_models.imagenet import data
    >>> import tensorflow_datasets as tfds
    >>> dataset = tfds.load('cifar10', split='train').batch(16)
    >>> dd = dataset.as_numpy_iterator().next()
    >>> images, labels = dd['image'], tf.one_hot(dd['label'], depth=10)
    >>> aa, bb = data.cutmix(images, labels)
    >>> print(bb.numpy()[bb.numpy() != 0])
    >>> plt.imshow(np.hstack(aa))
    """

    data.init_dataset.__doc__ = """ Init dataset by name.
    Args:
      data_name: the registered dataset name from `tensorflow_datasets`.
      input_shape: input shape.
      batch_size: batch size.
      buffer_size: dataset shuffle buffer size.
      info_only: boolean value if returns dataset info only.
      mixup_alpha: mixup applying probability.
      cutmix_alpha: cutmix applying probability.
      rescale_mode: one of ["tf", "torch", "raw01", "raw"]. Detail in `data.init_mean_std_by_rescale_mode`. Or specific `(mean, std)` like `(128.0, 128.0)`.
      eval_central_crop: central crop fraction for evaluation. Default `1.0` for disabling, < 0 values set a crop fraction.
      random_crop_min: min scale for `random_crop_fraction`. Max scale is `1.0`, Ratio is `(0.75, 1.3333333)`.
      resize_method: one of ["nearest", "bilinear", "bicubic"]. Resize method for `tf.image.resize`.
      resize_antialias: boolean value if using antialias for `tf.image.resize`.
      random_erasing_prob: if applying random erasing. Default 0 for disabling, > 0 values set probability.
      magnitude: randaug magnitude.
      num_layers: randaug num_layers.
      augment_kwargs: randaug kwargs. Too many to list them all.

    Returns: train_dataset, test_dataset, total_images, num_classes, steps_per_epoch
    """



from keras_cv_attention_models.aotnet import AotNet
from keras_cv_attention_models.download_and_load import reload_model_weights

PRETRAINED_DICT = {
    "resnet50d": {"imagenet": "1b71933a82b058ba1e605ee5c01f64b2"},
    "resnet101d": {"imagenet": "79b075be5cf222cff2bced7a5a117623"},
    "resnet152d": {"imagenet": "0a15299b9abe1fee3ae06d9a59d13a3f"},
    "resnet200d": {"imagenet": "b5961494e0072c342b838c77ef52ddc5"},
}


def ResNetD(num_blocks, input_shape=(224, 224, 3), pretrained="imagenet", stem_type="deep", strides=2, shortcut_type="avg", **kwargs):
    strides = strides if isinstance(strides, (list, tuple)) else [1, 2, 2, strides]
    model = AotNet(num_blocks, input_shape=input_shape, stem_type=stem_type, strides=strides, shortcut_type=shortcut_type, **kwargs)
    reload_model_weights(model, pretrained_dict=PRETRAINED_DICT, sub_release="resnet_family", pretrained=pretrained)
    return model


def ResNet50D(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 6, 3]
    return ResNetD(**locals(), model_name="resnet50d", **kwargs)


def ResNet101D(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 23, 3]
    return ResNetD(**locals(), model_name="resnet101d", **kwargs)


def ResNet152D(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 8, 36, 3]
    return ResNetD(**locals(), model_name="resnet152d", **kwargs)


def ResNet200D(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 24, 36, 3]
    return ResNetD(**locals(), model_name="resnet200d", **kwargs)



from keras_cv_attention_models.efficientformer.efficientformer import EfficientFormer, EfficientFormerL1, EfficientFormerL3, EfficientFormerL7
from keras_cv_attention_models.efficientformer.efficientformer_v2 import EfficientFormerV2, EfficientFormerV2S0, EfficientFormerV2S1, EfficientFormerV2S2, EfficientFormerV2L

__v1_head_doc__ = """
Keras implementation of [Github snap-research/efficientformer](https://github.com/snap-research/efficientformer).
Paper [PDF 2206.01191 EfficientFormer: Vision Transformers at MobileNet Speed](https://arxiv.org/pdf/2206.01191.pdf).
"""

__common_head_doc__ = """
Args:
  num_blocks: number of block for each stack.
  out_channels: output channels for each stack.
  mlp_ratios: int or list value indicates expand ratio for mlp blocks hidden channel in each stack.
      Each element can also be a int or list, indicates `mlp_ratio` for each block.
  num_attn_blocks_each_stack: int or list value indicates number of attention blocks in each stack tail.
      If int, means using same value for all stacks, or should be same length with `num_blocks`.
      `0` means not using attention block in relative stack, or should be less than relative `num_blocks`.
      Default `[0, 0, 0, 1]` for `EfficientFormer`, `[0, 0, 2, 2]` for `EfficientFormerV2`.
  stem_width: output dimension for stem block, default `-1` for using out_channels[0].
  stem_activation: activation for stem branch. None for same with `activation`
  model_name: string, model name.
"""

__tail_doc__ = """  layer_scale: layer scale init value. Default `-1` means not applying, any value `>=0` will add a scale value for each block output.
      [Going deeper with Image Transformers](https://arxiv.org/pdf/2103.17239.pdf).
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole stacks, default `gelu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be a constant value like `0.2`,
      or a tuple value like `(0, 0.2)` indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  dropout: top dropout rate if top layers is included. Default 0.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `None`.
  use_distillation: Boolean value if output `distill_head`. Default `True`.
  pretrained: one of `None` (random initialization) or 'imagenet' (pre-training on ImageNet).
      Will try to download and load pre-trained model weights if not None.
  **kwargs: other parameters if available.

Returns:
    A `keras.Model` instance.
"""

EfficientFormer.__doc__ = __v1_head_doc__ + __common_head_doc__ + __tail_doc__ + """
Model architectures:
  | Model                      | Params | FLOPs | Input | Top1 Acc |
  | -------------------------- | ------ | ----- | ----- | -------- |
  | EfficientFormerL1, distill | 12.3M  | 1.31G | 224   | 79.2     |
  | EfficientFormerL3, distill | 31.4M  | 3.95G | 224   | 82.4     |
  | EfficientFormerL7, distill | 74.4M  | 9.79G | 224   | 83.3     |
"""

EfficientFormerL1.__doc__ = __v1_head_doc__ + """
Args:
""" + __tail_doc__

EfficientFormerL3.__doc__ = EfficientFormerL1.__doc__
EfficientFormerL7.__doc__ = EfficientFormerL1.__doc__

__v2_head_doc__ = """
Keras implementation of [Github snap-research/efficientformer](https://github.com/snap-research/efficientformer).
Paper [PDF 2212.08059 Rethinking Vision Transformers for MobileNet Size and Speed](https://arxiv.org/pdf/2212.08059.pdf).
"""

EfficientFormerV2.__doc__ = __v2_head_doc__ + __common_head_doc__ + __tail_doc__ + """
Model architectures:
  | Model                        | Params | FLOPs  | Input | Top1 Acc |
  | ---------------------------- | ------ | ------ | ----- | -------- |
  | EfficientFormerV2S0, distill | 3.60M  | 405.2M | 224   | 76.2     |
  | EfficientFormerV2S1, distill | 6.19M  | 665.6M | 224   | 79.7     |
  | EfficientFormerV2S2, distill | 12.7M  | 1.27G  | 224   | 82.0     |
  | EfficientFormerV2L, distill  | 26.3M  | 2.59G  | 224   | 83.5     |
"""

EfficientFormerV2S0.__doc__ = __v2_head_doc__ + """
Args:
""" + __tail_doc__

EfficientFormerV2S1.__doc__ = EfficientFormerV2S0.__doc__
EfficientFormerV2S2.__doc__ = EfficientFormerV2S0.__doc__
EfficientFormerV2L.__doc__ = EfficientFormerV2S0.__doc__



import tensorflow as tf


@tf.keras.utils.register_keras_serializable(package="kecamLoss")
class BinaryCrossEntropyTimm(tf.keras.losses.BinaryCrossentropy):
    """
    >>> import torch, timm.loss
    >>> from keras_cv_attention_models.imagenet import losses
    >>> tt = timm.loss.BinaryCrossEntropy(smoothing=0.0, target_threshold=0.2)
    >>> ss = losses.BinaryCrossEntropyTimm(target_threshold=0.2, from_logits=True)
    >>> y_true = tf.one_hot(np.random.permutation(20).reshape(2, 10), 10).numpy()
    >>> y_true = np.clip(y_true[0] + y_true[1], 0, 1)
    >>> y_pred = np.random.uniform(size=(10, 10))
    >>> torch_out = tt(torch.from_numpy(y_pred), torch.from_numpy(y_true)).numpy()
    >>> keras_out = ss(y_true, y_pred).numpy()
    >>> print(f"{torch_out = }, {keras_out = }")
    # torch_out = array(0.9457581, dtype=float32), keras_out = 0.945758044719696
    """

    def __init__(self, target_threshold=0.0, label_smoothing=0.0, **kwargs):
        super().__init__(label_smoothing=label_smoothing, **kwargs)
        self.target_threshold = target_threshold
        self.label_smoothing = label_smoothing

    def call(self, y_true, y_pred):
        target_threshold = tf.cast(self.target_threshold, y_true.dtype)
        y_true = tf.where(y_true > target_threshold, tf.ones_like(y_true), tf.zeros_like(y_true))
        return super().call(y_true, y_pred)

    def get_config(self):
        config = super().get_config()
        config.update({"target_threshold": self.target_threshold, "label_smoothing": self.label_smoothing})
        return config


@tf.keras.utils.register_keras_serializable(package="kecamLoss")
class DistillKLDivergenceLoss(tf.keras.losses.Loss):
    """[PDF 2106.05237 Knowledge distillation: A good teacher is patient and consistent](https://arxiv.org/pdf/2106.05237.pdf)
    Modified according [Knowledge distillation recipes](https://keras.io/examples/keras_recipes/better_knowledge_distillation/)

    Temperature affecting:
    >>> teacher_prob = np.array([0, 0.2, 0.4, 0.6, 0.8, 1.0])
    >>> _ = [print("temperature:", temp, tf.nn.softmax(teacher_prob / temp).numpy()) for temp in [0.1, 1, 10, 20]]
    >>> # temperature: 0.1 [3.92559586e-05 2.90064480e-04 2.14330272e-03 1.58369840e-02 1.17020363e-01 8.64670029e-01]
    >>> # temperature: 1 [0.09542741 0.11655531 0.14236097 0.17388009 0.21237762 0.25939861]
    >>> # temperature: 10 [0.1584458  0.16164661 0.16491209 0.16824354 0.17164228 0.17510968]
    >>> # temperature: 20 [0.16252795 0.16416138 0.16581123 0.16747766 0.16916084 0.17086094]
    """

    def __init__(self, temperature=10, **kwargs):
        super().__init__(**kwargs)
        self.temperature = temperature
        # self.kl_divergence = tf.keras.losses.KLDivergence()

    def call(self, teacher_prob, student_prob):
        return tf.losses.kl_divergence(
            tf.nn.softmax(teacher_prob / self.temperature, axis=-1),
            tf.nn.softmax(student_prob / self.temperature, axis=-1),
        )


# Not using, from VOLO with mix_token lambda
def token_label_class_loss(y_true, y_pred):
    # tf.print(", y_true:", y_true.shape, "y_pred:", y_pred.shape, end="")
    if y_pred.shape[-1] != y_true.shape[-1]:
        y_pred, cls_lambda = y_pred[:, :-1], y_pred[:, -1:]
        y_true = tf.cast(y_true, y_pred.dtype)
        y_true = cls_lambda * y_true + (1 - cls_lambda) * y_true[::-1]
    return keras.losses.categorical_crossentropy(y_true, y_pred, from_logits=True)



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    batchnorm_with_activation,
    conv2d_no_bias,
    inverted_residual_block,
    make_divisible,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

PRETRAINED_DICT = {
    "fbnetv3_b": {"imagenet": "498af18b2241fa07e029fca393885d93"},
    "fbnetv3_d": {"imagenet": "086ddd9ccf516f3699e80f06f231d41c"},
    "fbnetv3_g": {"imagenet": "5ed06c04bfeeff798e2d951ec863214c"},
    "lcnet_050": {"imagenet": "f740b5d73a42f65c5f7baf436d75d75f", "ssld": "2bfa67fa829e799710d76880df485ceb"},
    "lcnet_075": {"imagenet": "c2dd36d7362a782e70d7b5c7b4422027"},
    "lcnet_100": {"imagenet": "cb349c3f11678abae2799b88163fc614", "ssld": "ac978d02c94be99e4c6c10c24abfc4a8"},
    "lcnet_150": {"imagenet": "c1d114d56e88ff67e5ae4222db64c4fb"},
    "lcnet_200": {"imagenet": "46a7426993ec2910b93eab479528af68"},
    "lcnet_250": {"imagenet": "6dbe4e7bd1dd7a3ec227e2140f2ea162", "ssld": "ca752dc4ee870201a9a12d2234f284b2"},
    "mobilenetv3_large_075": {"imagenet": "56dc3594efb27c1df2c679d5486b9379"},
    "mobilenetv3_large_100": {
        "imagenet": "e5dbb4947d1fac0e5b0f90c7b6c3b6e9",
        "miil_21k": "86d0e21b372ff02498062f55e253b61c",
        "miil": "b2cb65b167a16a54795d87f1aaf0bf1e",
    },
    "mobilenetv3_small_050": {"imagenet": "17063e82099f420d552cc5f06efd9b46"},
    "mobilenetv3_small_075": {"imagenet": "1df2126a9ed19704996d969f57afa7bb"},
    "mobilenetv3_small_100": {"imagenet": "aa84f2bb4d7faf9fe2417267d2fc35b1"},
    "tinynet_a": {"imagenet": "ccffe4208feb4e9573834ab9cd4074e8"},
    "tinynet_b": {"imagenet": "4888c408c2ac69bf7824da383c58d52d"},
    "tinynet_c": {"imagenet": "29ce15979b8800176621780b0ea91eaa"},
    "tinynet_d": {"imagenet": "e08615d88e71f1548d040e697926514e"},
    "tinynet_e": {"imagenet": "514fbbcd582db6d4297a140efb84af9a"},
}


def avg_pool_conv_output(inputs, output_num_features=1280, use_output_feature_bias=True, activation="hard_swish"):
    # nn = layers.AveragePooling2D(pool_size=inputs.shape[1:3], name="avg_pool")(inputs)
    # nn = layers.GlobalAveragePooling2D(name="avg_pool")(inputs)[:, None, None, :]
    h_axis, w_axis = [1, 2] if image_data_format() == "channels_last" else [2, 3]
    nn = functional.reduce_mean(inputs, [h_axis, w_axis], keepdims=True)  # using AvgPool2D cannot set dynamic input_shape
    if output_num_features > 0:
        nn = conv2d_no_bias(nn, make_divisible(output_num_features, 8), use_bias=use_output_feature_bias, name="features_")
        nn = activation_by_name(nn, activation, name="features_")
    nn = layers.Flatten()(nn)
    return nn


def conv_avg_pool_output(inputs, output_num_features=1280, use_output_feature_bias=True, activation="hard_swish"):
    nn = inputs
    if output_num_features > 0:
        nn = conv2d_no_bias(nn, make_divisible(output_num_features, 8), use_bias=use_output_feature_bias, name="features_")
        nn = batchnorm_with_activation(nn, activation=activation, name="features_")
    nn = layers.GlobalAveragePooling2D(name="avg_pool")(nn)
    return nn


def MobileNetV3(
    num_blocks=[1, 2, 3, 4, 2, 3],  # [Stack parameters]
    out_channels=[16, 24, 40, 80, 112, 160],
    expands=[1, [4, 3], 3, [6, 2.5, 2.3, 2.3], 6, 6],
    kernel_sizes=[3, 3, 5, 3, 3, 5],
    strides=[1, 2, 2, 2, 1, 2],
    activations=["relu", "relu", "relu", "hard_swish", "hard_swish", "hard_swish"],
    disable_shortcut=False,  # True for LCNet
    use_blocks_output_activation=False,  # True for LCNet
    width_ratio=1.0,
    stem_width=16,  # [Stem parameters]
    fix_stem=False,
    stem_feature_activation="hard_swish",  # "swish" for TinyNet
    se_ratios=[0, 0, 0.25, 0, 0.25, 0.25],  # [SE module parameters]
    se_activation=("relu", "hard_sigmoid_torch"),  # ("hard_swish", "hard_sigmoid_torch") for FBNetV3, None for TinyNet
    se_limit_round_down=0.9,  # 0.95 for FBNetV3
    se_divisor=8,  # 1 for TinyNet
    use_expanded_se_ratio=True,  # False for FBNetV3, TinyNet
    output_num_features=1280,  # [Output parameters]
    use_additional_output_conv=True,  # False for LCNet, TinyNet
    use_output_feature_bias=True,  # False for FBNetV3, TinyNet
    use_avg_pool_conv_output=True,  # False for TinyNet
    input_shape=(224, 224, 3),  # [Model common parameters]
    num_classes=1000,
    drop_connect_rate=0,
    classifier_activation="softmax",
    dropout=0,
    pretrained=None,
    model_name="mobilenetv3",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimention is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)
    stem_width = stem_width if fix_stem else make_divisible(stem_width * width_ratio, 8)
    nn = conv2d_no_bias(inputs, stem_width, kernel_size=3, strides=2, padding="same", name="stem_")
    nn = batchnorm_with_activation(nn, activation=stem_feature_activation, name="stem_")

    block_kwargs = {  # Common parameters for all blocks
        "is_torch_mode": True,
        "se_activation": se_activation,
        "se_divisor": se_divisor,
        "se_limit_round_down": se_limit_round_down,
    }

    """ stage [1, 2, 3, 4] """
    pre_out = stem_width
    global_block_id = 0
    total_blocks = sum(num_blocks)
    for id, (num_block, out_channel, kernel_size, stride) in enumerate(zip(num_blocks, out_channels, kernel_sizes, strides)):
        stage_name = "stack{}_".format(id + 1)
        out_channel = make_divisible(out_channel * width_ratio, 8)
        activation = activations[id] if isinstance(activations, (list, tuple)) else activations
        expand = expands[id] if isinstance(expands, (list, tuple)) else expands
        se_ratio = se_ratios[id] if isinstance(se_ratios, (list, tuple)) else se_ratios
        for block_id in range(num_block):
            name = stage_name + "block{}_".format(block_id + 1)
            stride = stride if block_id == 0 else 1
            shortcut = True if out_channel == pre_out and stride == 1 and not disable_shortcut else False
            cur_expand = expand[min(block_id, len(expand) - 1)] if isinstance(expand, (list, tuple)) else expand
            cur_se_ratio = se_ratio * cur_expand if use_expanded_se_ratio else se_ratio
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks

            nn = inverted_residual_block(
                nn, out_channel, stride, cur_expand, shortcut, kernel_size, block_drop_rate, cur_se_ratio, **block_kwargs, activation=activation, name=name
            )
            if use_blocks_output_activation:
                nn = activation_by_name(nn, activation, name=name + "output_")
            pre_out = out_channel
            global_block_id += 1

    if use_additional_output_conv:
        nn = conv2d_no_bias(nn, make_divisible(out_channels[-1] * cur_expand * width_ratio, 8), kernel_size=1, name="pre_output_")
        nn = batchnorm_with_activation(nn, activation=stem_feature_activation, name="pre_output_")

    if num_classes > 0:
        if use_avg_pool_conv_output:
            nn = avg_pool_conv_output(nn, output_num_features, use_output_feature_bias, stem_feature_activation)
        else:
            nn = conv_avg_pool_output(nn, output_num_features, use_output_feature_bias, stem_feature_activation)

        if dropout > 0:
            nn = layers.Dropout(dropout, name="head_drop")(nn)
        nn = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="predictions")(nn)
    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, PRETRAINED_DICT, "mobilenetv3_family", pretrained)
    return model


def MobileNetV3Large(model_name="mobilenetv3_large", **kwargs):
    kwargs.pop("kwargs", None)
    return MobileNetV3(**locals(), **kwargs)


def MobileNetV3Small(
    num_blocks=[1, 2, 3, 2, 3],
    out_channels=[16, 24, 40, 48, 96],
    expands=[1, [4.5, 3.67], [4, 6, 6], 3, 6],
    kernel_sizes=[3, 3, 5, 5, 5],
    strides=[2, 2, 2, 1, 2],
    activations=["relu", "relu", "hard_swish", "hard_swish", "hard_swish"],
    se_ratios=[0.25, 0, 0.25, 0.25, 0.25],
    output_num_features=1024,
    model_name="mobilenetv3_small",
    **kwargs,
):
    kwargs.pop("kwargs", None)
    return MobileNetV3(**locals(), **kwargs)


def MobileNetV3Large075(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return MobileNetV3Large(**locals(), width_ratio=0.75, model_name="mobilenetv3_large_075", **kwargs)


def MobileNetV3Large100(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return MobileNetV3Large(**locals(), model_name="mobilenetv3_large_100", **kwargs)


def MobileNetV3Small050(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    fix_stem = True
    return MobileNetV3Small(**locals(), width_ratio=0.5, model_name="mobilenetv3_small_050", **kwargs)


def MobileNetV3Small075(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return MobileNetV3Small(**locals(), width_ratio=0.75, model_name="mobilenetv3_small_075", **kwargs)


def MobileNetV3Small100(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return MobileNetV3Small(**locals(), model_name="mobilenetv3_small_100", **kwargs)



from keras_cv_attention_models.mobilevit.mobilevit import MobileViT, MobileViT_XXS, MobileViT_XS, MobileViT_S, linear_self_attention
from keras_cv_attention_models.mobilevit.mobilevit_v2 import (
    MobileViT_V2,
    MobileViT_V2_050,
    MobileViT_V2_075,
    MobileViT_V2_100,
    MobileViT_V2_125,
    MobileViT_V2_150,
    MobileViT_V2_175,
    MobileViT_V2_200,
)

__v1_head_doc__ = """
Keras implementation of [Github apple/ml-cvnets/mobilevit](https://github.com/apple/ml-cvnets/blob/main/cvnets/models/classification/mobilevit.py).
Paper [PDF 2110.02178 MOBILEVIT: LIGHT-WEIGHT, GENERAL-PURPOSE, AND MOBILE-FRIENDLY VISION TRANSFORMER](https://arxiv.org/pdf/2110.02178.pdf).
"""

__v2_head_doc__ = """
Keras implementation of [Github apple/ml-cvnets](https://github.com/apple/ml-cvnets).
Paper [PDF 2206.02680 Separable Self-attention for Mobile Vision Transformers](https://arxiv.org/pdf/2206.02680.pdf).
"""

__args__ = """
Args:
  num_blocks: number of blocks in each stack.
  out_channels: output channels for each stack.
  attn_channels: hidden channel for transformer blocks. Can be a list matching out_channels,
      or a float number for expansion ratio of out_channels.
  block_types: block types for each stack,
      - `conv` or any `c` / `C` starts word, means `bottle_in_linear_out_block` block.
      - `transfrom` or any not `c` / `C` starts word, means `mhsa_mlp_block` block.
      value could be in format like `"cctt"` or `"CCTT"` or `["conv", "conv", "transfrom", "transform"]`.
  strides: list value for stride in first block of each stack.
  expand_ratio: conv blocks hidden channel expand ratio. Default 2 for MobileViT_XXS, 4 for MobileViT_XS and MobileViT_S.
  stem_width: output dimension for stem block.
  patch_size: extracted patch size for transformer blocks.
  patches_to_batch: boolean value if stack extracted patches to batch dimension. True for V1, False for V2.
  resize_first: boolena value if perform `resize -> conv` or `conv -> resize` if input shape not divisible by 32.
      False for V1, True for V2.
  use_depthwise: boolean value if using depthwise_conv2d or conv2d for `transformer_pre_process` process.
      False for V1, True for V2.
  use_fusion: boolean value if fusing with tensor before attention blocks in `transformer_post_process` process.
      True for V1, False for V2.
  num_norm_groups: norm layer froups number. -1 or 0 for V1 using layer_norm, or 1 for V2 using group_norm.
  use_linear_attention: boolean value if using `linear_self_attention` or `multi_head_self_attention` in attention block.
      False for V1, True for V2.
  use_conv_mlp: boolean value if using `conv` or `dense` for mlp blocks. False for V1, True for V2.
  output_num_features: none `0` value to add another `conv2d + bn + activation` layers before `GlobalAveragePooling2D`.
  model_name: string, model name.
"""

__tail_doc__ = """  layer_scale: layer scale init value, [Going deeper with Image Transformers](https://arxiv.org/abs/2103.17239).
      Default `-1` for not using.
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `relu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
  dropout: dropout rate if top layers is included.
  pretrained: None or {pretrained}.
  **kwargs: other parameters if available.

Returns:
    A `keras.Model` instance.
"""

MobileViT.__doc__ = __v1_head_doc__ + __args__ + __tail_doc__.format(pretrained="imagenet") + """
Model architectures:
  | Model         | Params | FLOPs | Input | Top1 Acc |
  | ------------- | ------ | ----- | ----- | -------- |
  | MobileViT_XXS | 1.3M   | 0.42G | 256   | 69.0     |
  | MobileViT_XS  | 2.3M   | 1.05G | 256   | 74.7     |
  | MobileViT_S   | 5.6M   | 2.03G | 256   | 78.3     |
"""

MobileViT_V2.__doc__ = __v2_head_doc__ + __args__ + __tail_doc__.format(pretrained="one of [imagenet, imagenet22k]") + """
Model architectures:
  | Model              | Params | FLOPs | Input | Top1 Acc |
  | ------------------ | ------ | ----- | ----- | -------- |
  | MobileViT_V2_050   | 1.37M  | 0.47G | 256   | 70.18    |
  | MobileViT_V2_075   | 2.87M  | 1.04G | 256   | 75.56    |
  | MobileViT_V2_100   | 4.90M  | 1.83G | 256   | 78.09    |
  | MobileViT_V2_125   | 7.48M  | 2.84G | 256   | 79.65    |
  | MobileViT_V2_150   | 10.6M  | 4.07G | 256   | 80.38    |
  | - imagenet22k      | 10.6M  | 4.07G | 256   | 81.46    |
  | - imagenet22k, 384 | 10.6M  | 9.15G | 384   | 82.60    |
  | MobileViT_V2_175   | 14.3M  | 5.52G | 256   | 80.84    |
  | - imagenet22k      | 14.3M  | 5.52G | 256   | 81.94    |
  | - imagenet22k, 384 | 14.3M  | 12.4G | 384   | 82.93    |
  | MobileViT_V2_200   | 18.4M  | 7.12G | 256   | 81.17    |
  | - imagenet22k      | 18.4M  | 7.12G | 256   | 82.36    |
  | - imagenet22k, 384 | 18.4M  | 16.2G | 384   | 83.41    |
"""

MobileViT_XXS.__doc__ = __v1_head_doc__ + """
Args:
""" + __tail_doc__.format(pretrained="imagenet")

MobileViT_XS.__doc__ = MobileViT_XXS.__doc__
MobileViT_S.__doc__ = MobileViT_XXS.__doc__

MobileViT_V2_050.__doc__ = __v2_head_doc__ + """
Args:
""" + __tail_doc__.format(pretrained="one of [imagenet, imagenet22k]")

MobileViT_V2_075.__doc__ = MobileViT_V2_050.__doc__
MobileViT_V2_100.__doc__ = MobileViT_V2_050.__doc__
MobileViT_V2_125.__doc__ = MobileViT_V2_050.__doc__
MobileViT_V2_150.__doc__ = MobileViT_V2_050.__doc__
MobileViT_V2_175.__doc__ = MobileViT_V2_050.__doc__
MobileViT_V2_200.__doc__ = MobileViT_V2_050.__doc__

linear_self_attention.__doc__ = __v2_head_doc__ + """
Linear Self Attention. Defined as function, not layer.

Args:
  inputs: input tensor.
  qkv_bias: Boolean, whether the qkv dense layer use bias vectors/matrices.
  out_bias: Boolean, whether the ouput dense layer use bias vectors/matrices.
  attn_axis: axis for applying attention softmax operation. Default 2 for MobileViT_V2 applying on `patch_hh * patch_ww` dimension.
  attn_dropout: Dropout probability for attention.

Examples:
>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14, 16, 256])
>>> nn = attention_layers.linear_self_attention(inputs, attn_axis=-1)
>>> print(f"{nn.shape = }")
# nn.shape = TensorShape([None, 14, 16, 512])

>>> mm = keras.models.Model(inputs, nn)
>>> mm.summary()
>>> print({ii.name: ii.shape for ii in mm.weights})
# {'conv2d_2/kernel:0': TensorShape([1, 1, 256, 513]),
#  'conv2d_3/kernel:0': TensorShape([1, 1, 256, 256])}
"""



from keras_cv_attention_models.aotnet.aotnet import (
    AotNet,
    AotNet50,
    AotNet101,
    AotNet152,
    AotNet200,
    AotNetV2,
    AotNet50V2,
    AotNet101V2,
    AotNet152V2,
    AotNet200V2,
    DEFAULT_PARAMS,
)

__head_doc__ = """
AotNet is just a `ResNet` / `ResNetV2` like framework.
Set parameters like `attn_types` and `attn_params` and others, which is used to apply different types attention layers.
Default parameters set is a typical `ResNet` architecture with `Conv2D use_bias=False` and `padding` like `PyTorch`.
"""

__tail_doc__ = """
  [Stack parameters]
  preact: whether to use pre-activation or not. Default `False` for ResNet like, `True` for ResNetV2 like.
  strides: a `number` or `list`, indicates strides used in the last stack or list value for all stacks.
      If a number, for `ResNet` like models, it will be `[1, 2, 2, strides]`.
      If a number, for `ResNetV2` like models, it will be `[2, 2, 2, strides]`.
      Default [1, 2, 2, 2]
  strides_first: boolean value if use downsample in the first block in each stack.
      Default `True` for ResNet like, `False` for ResNetV2 like.
  out_channels: default as `[64, 128, 256, 512]`. Output channel for each stack.
  hidden_channel_ratio: filter expansion for each block hidden layers. The larger the wider. For default `ResNet` / `ResNetV2` like, it's `0.25`.
  use_3x3_kernel: boolean value if use two `3x3` conv instead of `1x1 + 3x3 + 1x1` conv in each block. Default `False`.
  use_block_output_activation: boolean value if add activation after block `Add` layer. Default `True`.

  [Stem parameters]
  stem_width: output dimension for stem block. Default 64.
  stem_type: string in value `[None, "deep", "quad", "tiered", "kernel_3x3"]`. Indicates diffrerent stem type. Default None.
  quad_stem_act: boolean value if add `BN + act` after first `Conv` layer for `stem_type="quad"`. Default `False`.
  stem_last_strides: the last strides in stem block. Default `1`
  stem_downsample: boolean value if add `MaxPooling2D` layer after stem block. Default `True`.

  [Attention block parameters]
  attn_types: is a `string` or `list`, indicates attention layer type for each stack.
      Each element can also be a `string` or `list`, indicates attention layer type for each block.
      - `"bot"`: `mhsa_with_relative_position_embedding` from `botnet`. Default values: `num_heads=4, relative=True, out_bias=False`.
      - `"cot"`: `cot_attention` from `cotnet`. Default values: `kernel_size=3, downsample_first=True`.
      - `"halo"`: `halo_attention` from `halonet`. Default values: `num_heads=8, block_size=4, halo_size=1`.
      - `"outlook"`: `outlook_attention` from `volo`. Default values: `num_head=8, kernel_size=3`.
      - `"sa"`: `split_attention_conv2d` from `resnest`. Default values: `kernel_size=3, groups=2, downsample_first=False`.
      - `None`: `Conv2D`. Can add `groups` like `ResNeXt` or add `se` and `eca` attention.
  attn_params: like `attn_types`, is a dict or list, each element in list can also be a dict or list.
      Indicates specific attention layer parameters for relative `attn_types`.
  se_ratio: value in `(0, 1)`, where `0` means not using `se_module`. Should be a `number` or `list`, indicates `se_ratio` for each stack.
      Each element can also be a `number` or `list`, indicates `se_ratio` for each block.
  use_eca: boolean value if use `eca` attention. Can also be a list like `se_ratio`.
  groups: `groups` for `Conv2D` layer if relative `attn_types` is `None`. `ResNeXt` like archeticture. Note it's NOT the `group_size`.
      Default value `1` means not using group.
  group_size: group size for `Conv2D` layer if relative `attn_types` is `None`. `groups = filters / group_size`.
      Priority is higher than `groups`. Default `0` means not using group.
  bn_after_attn: boolean value if add `batchnorm + activation` layers after attention layer. Default `True`.

  [Shortcut branch parameters]
  shortcut_type: value in `["conv", "avg", "anti_alias", None]`. Indicates shortcut branch type if should apply `conv_shortcut`.
      - "conv": basic `Resnet` like conv branch.
      - "avg": `AvgPool2D + Conv2D`.
      - "anti_alias": `anti_alias_downsample + Conv2D`.
      - None: no shortcut.
      Default is `"conv"`.

  [Model common parameters]
  input_shape: it should have exactly 3 inputs channels, default `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `relu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `softmax`.
  output_num_features: none `0` value to add another `conv2d + bn + activation` layers before `GlobalAveragePooling2D`.
  dropout: dropout rate if top layers is included.
  pretrained: mostly it's `None` or `"imagenet"` if any.
  kwargs: Not used, only recieving parameter.

Returns:
    A `keras.Model` instance.
"""

AotNet.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  model_name: string, model name.
""" + __tail_doc__ + """
Example:

Definition of `BotNet26T`
>>> from keras_cv_attention_models import aotnet
>>> model = aotnet.AotNet(
>>>     num_blocks=[2, 2, 2, 2],
>>>     attn_types=[None, None, [None, "bot"], "bot"],
>>>     attn_params={"num_heads": 4, "out_weight": False},
>>>     stem_type="tiered",
>>>     input_shape=(256, 256, 3),
>>>     model_name="botnet26t",
>>> )
>>> model.summary()

Definition of `HaloNet50T`
>>> from keras_cv_attention_models import aotnet
>>> attn_params = [
>>>     None,
>>>     [None, None, None, {"block_size": 8, "halo_size": 3, "num_heads": 4, "out_weight": False}],
>>>     [None, {"block_size": 8, "halo_size": 3, "num_heads": 8, "out_weight": False}] * 3,
>>>     [None, {"block_size": 8, "halo_size": 3, "num_heads": 8, "out_weight": False}, None],
>>> ]
>>> model = aotnet.AotNet(
>>>     num_blocks=[3, 4, 6, 3],
>>>     attn_types=[None, [None, None, None, "halo"], [None, "halo"] * 3, [None, "halo", None]],
>>>     attn_params=attn_params,
>>>     stem_type="tiered",
>>>     input_shape=(256, 256, 3),
>>>     model_name="halonet50t",
>>> )
>>> model.summary()

Mixing se and outlook and halo and bot and cot, 21M parameters
>>> # 50 is just a picked number that larger than the relative `num_block`
>>> model = aotnet.AotNet50V2(
>>>     attn_types=[None, "outlook", ["bot", "halo"] * 50, "cot"],
>>>     se_ratio=[0.25, 0, 0, 0],
>>>     stem_type="deep",
>>>     strides=1,
>>> )
>>> model.summary()
"""

AotNet50.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

AotNet101.__doc__ = AotNet50.__doc__
AotNet152.__doc__ = AotNet50.__doc__

AotNetV2.__doc__ = AotNet.__doc__
AotNet50V2.__doc__ = AotNet50.__doc__
AotNet101V2.__doc__ = AotNet50.__doc__
AotNet152V2.__doc__ = AotNet50.__doc__



import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format
from keras_cv_attention_models.aotnet import AotNet
from keras_cv_attention_models.download_and_load import reload_model_weights
from keras_cv_attention_models.attention_layers import batchnorm_with_activation, conv2d_no_bias, CompatibleExtractPatches, group_norm

BATCH_NORM_EPSILON = 1e-5
PRETRAINED_DICT = {
    "cotnet101": {"imagenet": {224: "589d2c817699d96085136f2af2dd2036"}},
    "cotnet50": {"imagenet": {224: "e2b1fd313834deebb1c1f83451525254"}},
    "cotnet_se101d": {"imagenet": {224: "4e2fc28b9a92259269f1b5544e67ab50"}},
    "cotnet_se152d": {"imagenet": {224: "6a2744af16b8cc4177fef52aba7ff083", 320: "9dad11a2ec3d2c8ecac9832fcf1e9ad3"}},
    "cotnet_se50d": {"imagenet": {224: "d1e40b172d26925794f0c9dea090dba7"}},
}


def cot_attention(inputs, kernel_size=3, strides=1, downsample_first=True, activation="relu", name=None):
    if downsample_first and strides > 1:
        inputs = layers.ZeroPadding2D(padding=1, name=name and name + "pool_pad")(inputs)
        inputs = layers.AvgPool2D(3, strides=2, name=name and name + "pool")(inputs)

    # inputs, kernel_size, strides, activation, name = tf.ones([1, 7, 7, 512]), 3, 1, "relu", ""
    height_axis, width_axis, channel_axis = (1, 2, 3) if image_data_format() == "channels_last" else (2, 3, 1)
    filters = inputs.shape[channel_axis]
    randix = 2

    # key_embed
    if kernel_size // 2 != 0:
        key_input = layers.ZeroPadding2D(padding=kernel_size // 2, name=name and name + "conv_pad")(inputs)
    else:
        key_input = inputs
    key = conv2d_no_bias(key_input, filters, kernel_size, groups=4, name=name and name + "key_")
    key = batchnorm_with_activation(key, activation=activation, zero_gamma=False, name=name and name + "key_")

    # query key
    qk = layers.Concatenate(axis=channel_axis)([inputs, key])
    height, width = qk.shape[height_axis], qk.shape[width_axis]

    # embed weights from query and key, ignore `num_heads`, as it's set as `1`
    reduction = 8
    embed_ww = conv2d_no_bias(qk, filters // randix, 1, name=name and name + "embed_ww_1_")
    embed_ww = batchnorm_with_activation(embed_ww, activation=activation, zero_gamma=False, name=name and name + "embed_ww_1_")
    embed_filters = kernel_size * kernel_size * filters // reduction
    embed_ww = conv2d_no_bias(embed_ww, embed_filters, 1, use_bias=True, name=name and name + "embed_ww_2_")
    embed_ww = group_norm(embed_ww, groups=filters // reduction, epsilon=BATCH_NORM_EPSILON, name=name and name + "embed_ww_")

    # matmul, local_conv
    embed = conv2d_no_bias(inputs, filters, 1, name=name and name + "embed_1_")
    embed = batchnorm_with_activation(embed, activation=None, zero_gamma=False, name=name and name + "embed_1_")

    # unfold_j = torch.nn.Unfold(kernel_size=kernel_size, dilation=1, padding=1, stride=1)
    # x2 = unfold_j(bb).view(-1, reduction, filters // reduction, kernel_size * kernel_size, height, width)
    # y2 = (ww.unsqueeze(2) * x2.unsqueeze(1)).sum(-3).view(-1, filters, height, width)
    # sizes, patch_strides = [1, kernel_size, kernel_size, 1], [1, 1, 1, 1]
    # embed = layers.ZeroPadding2D(padding=kernel_size // 2, name=name and name + "embed_pad")(embed)
    # embed = functional.extract_patches(embed, sizes=sizes, strides=patch_strides, rates=(1, 1, 1, 1), padding="VALID")
    embed = CompatibleExtractPatches(sizes=kernel_size, strides=1, name=name and name + "patchs_")(embed)

    if image_data_format() == "channels_last":
        embed_ww = functional.reshape(embed_ww, (-1, height, width, filters // reduction, kernel_size * kernel_size))
        reduction_axis, kernel_axis = -2, -3
        embed_ww = functional.expand_dims(functional.transpose(embed_ww, [0, 1, 2, 4, 3]), axis=reduction_axis)  # expand dim on `reduction` axis
        embed = functional.reshape(embed, [-1, height, width, kernel_size * kernel_size, reduction, filters // reduction])
    else:
        embed_ww = functional.reshape(embed_ww, (-1, filters // reduction, kernel_size * kernel_size, height, width))
        reduction_axis, kernel_axis = 2, 1
        embed_ww = functional.expand_dims(functional.transpose(embed_ww, [0, 2, 1, 3, 4]), axis=reduction_axis)  # expand dim on `reduction` axis
        embed = functional.reshape(embed, [-1, kernel_size * kernel_size, reduction, filters // reduction, height, width])

    embed_out = layers.Multiply(name=name and name + "local_conv_mul")([embed, embed_ww])
    embed_out = functional.reduce_sum(embed_out, axis=kernel_axis)  # reduce on `kernel_size * kernel_size` axis
    embed_out = functional.reshape(embed_out, [-1, height, width, filters] if image_data_format() == "channels_last" else [-1, filters, height, width])
    embed_out = batchnorm_with_activation(embed_out, activation="swish", zero_gamma=False, name=name and name + "embed_2_")

    # attention
    attn = layers.Add()([embed_out, key])
    attn = functional.reduce_mean(attn, axis=[height_axis, width_axis], keepdims=True)
    # attn se module
    attn_se_filters = max(filters * randix // 4, 32)
    # attn = layers.Dense(attn_se_filters, use_bias=True, name=name and name + "attn_se_dense_1")(attn)
    attn = conv2d_no_bias(attn, attn_se_filters, 1, use_bias=True, name=name and name + "attn_se_1_")
    attn = batchnorm_with_activation(attn, activation=activation, zero_gamma=False, name=name and name + "attn_se_")
    # attn = layers.Dense(filters * randix, use_bias=True, name=name and name + "attn_se_dense_2")(attn)
    attn = conv2d_no_bias(attn, filters * randix, 1, use_bias=True, name=name and name + "attn_se_2_")
    attn = functional.reshape(attn, [-1, 1, 1, filters, randix] if image_data_format() == "channels_last" else [-1, filters, randix, 1, 1])
    # attn = functional.nn.softmax(attn, axis=-1)
    randix_axis = -1 if image_data_format() == "channels_last" else 2
    attn = layers.Softmax(axis=randix_axis, name=name and name + "attention_scores")(attn)

    # value and output
    value = layers.Concatenate(axis=randix_axis)([functional.expand_dims(embed_out, randix_axis), functional.expand_dims(key, randix_axis)])
    output = layers.Multiply()([value, attn])
    output = functional.reduce_sum(output, axis=randix_axis, name=name and name + "out")

    if not downsample_first and strides > 1:
        output = layers.ZeroPadding2D(padding=1, name=name and name + "pool_pad")(output)
        output = layers.AvgPool2D(3, strides=2, name=name and name + "pool")(output)
    return output


def CotNet(input_shape=(224, 224, 3), bn_after_attn=False, shortcut_type="avg", attn_types="cot", pretrained="imagenet", **kwargs):
    model = AotNet(input_shape=input_shape, attn_types=attn_types, bn_after_attn=bn_after_attn, shortcut_type=shortcut_type, **kwargs)
    reload_model_weights(model, pretrained_dict=PRETRAINED_DICT, sub_release="cotnet", pretrained=pretrained)
    return model


def CotNet50(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 6, 3]
    strides = [1, 2, 2, 2]
    return CotNet(**locals(), **kwargs, model_name="cotnet50")


def CotNet101(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 23, 3]
    strides = [1, 2, 2, 2]
    return CotNet(**locals(), **kwargs, model_name="cotnet101")


def CotNetSE50D(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 6, 3]
    strides = [2, 2, 2, 2]
    attn_types = ["sa", "sa", ["cot", "sa"] * 3, "cot"]
    attn_params = [
        {"downsample_first": True, "groups": 1, "activation": "swish"},  # sa, stack1
        {"downsample_first": True, "groups": 1, "activation": "swish"},  # sa, stack2
        [{}, {"downsample_first": True, "groups": 1, "activation": "swish"}] * 3,  # cot + sa, stack 3
        {},  # cot, stack 4
    ]
    stem_type = "deep"
    stem_width = 64
    stem_downsample = False
    return CotNet(**locals(), **kwargs, model_name="cotnet_se50d")


def CotNetSE101D(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 23, 3]
    strides = [2, 2, 2, 2]
    attn_types = ["sa", "sa", ["cot", "sa"] * 12, "cot"]
    attn_params = [
        {"downsample_first": True, "groups": 1, "activation": "swish"},  # sa, stack1
        {"downsample_first": True, "groups": 1, "activation": "swish"},  # sa, stack2
        [{}, {"downsample_first": True, "groups": 1, "activation": "swish"}] * 12,  # cot + sa, stack 3
        {},  # cot, stack 4
    ]
    stem_type = "deep"
    stem_width = 128
    stem_downsample = False
    return CotNet(**locals(), **kwargs, model_name="cotnet_se101d")


def CotNetSE152D(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 8, 36, 3]
    strides = [2, 2, 2, 2]
    attn_types = ["sa", "sa", ["cot", "sa"] * 18, "cot"]
    attn_params = [
        {"downsample_first": True, "groups": 1, "activation": "swish"},  # sa, stack1
        {"downsample_first": True, "groups": 1, "activation": "swish"},  # sa, stack2
        [{"downsample_first": False}, {"downsample_first": True, "groups": 1, "activation": "swish"}] * 18,  # cot + sa, stack 3
        {"downsample_first": False},  # cot, stack 4
    ]
    stem_type = "deep"
    stem_width = 128
    stem_downsample = False
    return CotNet(**locals(), **kwargs, model_name="cotnet_se152d")



from keras_cv_attention_models.ghostnet.ghostnet_v2 import GhostNetV2


def GhostNet(
    kernel_sizes=[3, 3, 3, 5, 5, 3, 3, 3, 3, 3, 3, 5, 5, 5, 5, 5],
    first_ghost_channels=[16, 48, 72, 72, 120, 240, 200, 184, 184, 480, 672, 672, 960, 960, 960, 960],
    out_channels=[16, 24, 24, 40, 40, 80, 80, 80, 80, 112, 112, 160, 160, 160, 160, 160],
    se_ratios=[0, 0, 0, 0.25, 0.25, 0, 0, 0, 0, 0.25, 0.25, 0.25, 0, 0.25, 0, 0.25],
    strides=[1, 2, 1, 2, 1, 2, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1],
    stem_width=16,
    stem_strides=2,
    width_mul=1.0,
    num_ghost_module_v1_stacks=-1,  # num of `ghost_module` stcks on the head, others are `ghost_module_multiply`, set `-1` for all using `ghost_module`
    output_conv_filter=-1,  # -1 for first_ghost_channels[-1] * width_mul
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="relu",
    classifier_activation="softmax",
    dropout=0,
    pretrained=None,
    model_name="ghostnet",
    kwargs=None,
):
    return GhostNetV2(**locals())


def GhostNet_050(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return GhostNet(**locals(), width_mul=0.5, model_name="ghostnet_050", **kwargs)


def GhostNet_100(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return GhostNet(**locals(), model_name="ghostnet_100", **kwargs)


def GhostNet_130(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return GhostNet(**locals(), width_mul=1.3, model_name="ghostnet_130", **kwargs)



import math
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    batchnorm_with_activation,
    conv2d_no_bias,
    depthwise_conv2d_no_bias,
    make_divisible,
    se_module,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

PRETRAINED_DICT = {
    "ghostnetv2_100": {"imagenet": "4f28597d5f72731ed4ef4f69ec9c1799"},
    "ghostnet_050": {"imagenet": "dbb5a89e19fa78f2f35f38b4c1ae4351"},
    "ghostnet_100": {"imagenet": "19a0f0f03f20e4bd6c1736102b4d979d"},
    "ghostnet_130": {"imagenet": "3a73bc721765c516a894b567674fc60b", "ssld": "62571bb90d71a7487679ae97642d13fb"},
}


def decoupled_fully_connected_attention_block(inputs, out_channel, name=""):
    # nn = layers.AvgPool2D(pool_size=2, strides=2, padding="SAME")(inputs)
    nn = layers.AvgPool2D(pool_size=2, strides=2)(inputs)
    nn = conv2d_no_bias(nn, out_channel, name=name + "1_")
    nn = batchnorm_with_activation(nn, activation=None, name=name + "1_")
    nn = depthwise_conv2d_no_bias(nn, (1, 5), padding="SAME", name=name + "2_")
    nn = batchnorm_with_activation(nn, activation=None, name=name + "2_")
    nn = depthwise_conv2d_no_bias(nn, (5, 1), padding="SAME", name=name + "3_")
    nn = batchnorm_with_activation(nn, activation=None, name=name + "3_")
    nn = activation_by_name(nn, "sigmoid", name=name)
    # print(f"{nn.shape = }, {nn.shape = }")
    size = functional.shape(inputs)[1:-1] if image_data_format() == "channels_last" else functional.shape(inputs)[2:]  # For dynamic shape
    nn = functional.resize(nn, size, antialias=False, method="bilinear")
    # nn = layers.UpSampling2D(size=(2, 2), interpolation='bilinear')(nn)
    # if int(nn.shape[1] - nn.shape[1]) > 0 or int(nn.shape[2] - nn.shape[2]) > 0:
    #     nn = nn[:, :nn.shape[1], :nn.shape[2]]
    # nn = layers.Resizing(nn.shape[1], nn.shape[2])(nn)
    return nn


def ghost_module(inputs, out_channel, use_dfc_block=False, activation="relu", name=""):
    ratio = 2
    hidden_channel = int(math.ceil(float(out_channel) / ratio))
    # print("[ghost_module] out_channel:", out_channel, "conv_out_channel:", conv_out_channel)
    primary_conv = conv2d_no_bias(inputs, hidden_channel, name=name + "prim_")
    primary_conv = batchnorm_with_activation(primary_conv, activation=activation, name=name + "prim_")

    # hidden_channel_cheap = int(out_channel - hidden_channel_prim)
    # cheap_conv = conv2d_no_bias(primary_conv, hidden_channel_cheap, kernel_size=3, padding="SAME", groups=hidden_channel_prim, name=name + "cheap_")
    cheap_conv = depthwise_conv2d_no_bias(primary_conv, kernel_size=3, padding="SAME", name=name + "cheap_")
    cheap_conv = batchnorm_with_activation(cheap_conv, activation=activation, name=name + "cheap_")
    ghost_out = layers.Concatenate()([primary_conv, cheap_conv])

    if use_dfc_block:
        shortcut = decoupled_fully_connected_attention_block(inputs, out_channel, name=name + "short_")
        ghost_out = layers.Multiply()([shortcut, ghost_out])
    return ghost_out


def ghost_bottleneck(
    inputs, out_channel, first_ghost_channel, kernel_size=3, strides=1, se_ratio=0, shortcut=True, use_dfc_block=False, activation="relu", name=""
):
    if shortcut:
        shortcut = depthwise_conv2d_no_bias(inputs, kernel_size, strides, padding="same", name=name + "short_1_")
        shortcut = batchnorm_with_activation(shortcut, activation=None, name=name + "short_1_")
        shortcut = conv2d_no_bias(shortcut, out_channel, name=name + "short_2_")
        shortcut = batchnorm_with_activation(shortcut, activation=None, name=name + "short_2_")
    else:
        shortcut = inputs

    nn = ghost_module(inputs, first_ghost_channel, use_dfc_block=use_dfc_block, activation=activation, name=name + "ghost_1_")

    if strides > 1:
        nn = depthwise_conv2d_no_bias(nn, kernel_size, strides, padding="same", name=name + "down_")
        nn = batchnorm_with_activation(nn, activation=None, name=name + "down_")

    if se_ratio > 0:
        nn = se_module(nn, se_ratio=se_ratio, divisor=4, activation=(activation, "hard_sigmoid_torch"), name=name + "se_")

    nn = ghost_module(nn, out_channel, activation=None, name=name + "ghost_2_")
    # print(f">>>> {strides = }, {inputs.shape = }, {shortcut.shape = }, {nn.shape = }")
    return layers.Add(name=name + "output")([shortcut, nn])


def GhostNetV2(
    kernel_sizes=[3, 3, 3, 5, 5, 3, 3, 3, 3, 3, 3, 5, 5, 5, 5, 5],
    first_ghost_channels=[16, 48, 72, 72, 120, 240, 200, 184, 184, 480, 672, 672, 960, 960, 960, 960],
    out_channels=[16, 24, 24, 40, 40, 80, 80, 80, 80, 112, 112, 160, 160, 160, 160, 160],
    se_ratios=[0, 0, 0, 0.25, 0.25, 0, 0, 0, 0, 0.25, 0.25, 0.25, 0, 0.25, 0, 0.25],
    strides=[1, 2, 1, 2, 1, 2, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1],
    stem_width=16,
    stem_strides=2,
    width_mul=1.0,
    num_ghost_module_v1_stacks=2,  # num of `ghost_module` stcks on the head, others are `ghost_module` with `dfc_block`, set `-1` for all using `ghost_module`
    output_conv_filter=-1,  # -1 for first_ghost_channels[-1] * width_mul
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="relu",
    classifier_activation="softmax",
    dropout=0,
    pretrained=None,
    model_name="ghostnetv2",
    kwargs=None,
):
    # Regard input_shape as force using original shape if first element is None or -1,
    # else assume channel dimention is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)
    channel_axis = -1 if image_data_format() == "channels_last" else 1

    stem_width = make_divisible(stem_width * width_mul, divisor=4)
    nn = conv2d_no_bias(inputs, stem_width, 3, strides=stem_strides, padding="same", name="stem_")
    nn = batchnorm_with_activation(nn, activation=activation, name="stem_")

    """ stages """
    for stack_id, (kernel, stride, first_ghost, out_channel, se_ratio) in enumerate(zip(kernel_sizes, strides, first_ghost_channels, out_channels, se_ratios)):
        stack_name = "stack{}_".format(stack_id + 1)
        out_channel = make_divisible(out_channel * width_mul, 4)
        first_ghost_channel = make_divisible(first_ghost * width_mul, 4)
        shortcut = False if out_channel == nn.shape[channel_axis] and stride == 1 else True
        use_dfc_block = True if num_ghost_module_v1_stacks >= 0 and stack_id >= num_ghost_module_v1_stacks else False
        nn = ghost_bottleneck(nn, out_channel, first_ghost_channel, kernel, stride, se_ratio, shortcut, use_dfc_block, activation=activation, name=stack_name)

    output_conv_filter = output_conv_filter if output_conv_filter > 0 else make_divisible(first_ghost_channels[-1] * width_mul, 4)
    nn = conv2d_no_bias(nn, output_conv_filter, 1, strides=1, name="pre_")
    nn = batchnorm_with_activation(nn, activation=activation, name="pre_")

    if num_classes > 0:
        nn = layers.GlobalAveragePooling2D(keepdims=True)(nn)
        nn = conv2d_no_bias(nn, 1280, 1, strides=1, use_bias=True, name="features_")
        nn = activation_by_name(nn, activation, name="features_")
        nn = layers.Flatten()(nn)
        if dropout > 0 and dropout < 1:
            nn = layers.Dropout(dropout)(nn)
        nn = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="head")(nn)

    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, PRETRAINED_DICT, "ghostnetv2", pretrained)
    return model


def GhostNetV2_100(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return GhostNetV2(**locals(), model_name="ghostnetv2_100", **kwargs)



from keras_cv_attention_models.convnext.convnext import ConvNeXt


def ConvNeXtV2(
    num_blocks=[3, 3, 9, 3],
    out_channels=[96, 192, 384, 768],
    stem_width=-1,
    layer_scale_init_value=0,  # 1e-6 for v1, 0 for v2
    use_grn=True,  # False for v1, True for v2
    head_init_scale=1.0,
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu",
    drop_connect_rate=0.1,
    classifier_activation="softmax",
    dropout=0,
    pretrained=None,
    model_name="convnext_v2",
    kwargs=None,
):
    return ConvNeXt(**locals())


def ConvNeXtV2Atto(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 2, 6, 2]
    out_channels = [40, 80, 160, 320]
    return ConvNeXtV2(**locals(), model_name="convnext_v2_atto", **kwargs)


def ConvNeXtV2Femto(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 2, 6, 2]
    out_channels = [48, 96, 192, 384]
    return ConvNeXtV2(**locals(), model_name="convnext_v2_femto", **kwargs)


def ConvNeXtV2Pico(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 2, 6, 2]
    out_channels = [64, 128, 256, 512]
    return ConvNeXtV2(**locals(), model_name="convnext_v2_pico", **kwargs)


def ConvNeXtV2Nano(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 2, 8, 2]
    out_channels = [80, 160, 320, 640]
    return ConvNeXtV2(**locals(), model_name="convnext_v2_nano", **kwargs)


def ConvNeXtV2Tiny(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 9, 3]
    out_channels = [96, 192, 384, 768]
    return ConvNeXtV2(**locals(), model_name="convnext_v2_tiny", **kwargs)


def ConvNeXtV2Base(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 27, 3]
    out_channels = [128, 256, 512, 1024]
    return ConvNeXtV2(**locals(), model_name="convnext_v2_base", **kwargs)


def ConvNeXtV2Large(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 27, 3]
    out_channels = [192, 384, 768, 1536]
    return ConvNeXtV2(**locals(), model_name="convnext_v2_large", **kwargs)


def ConvNeXtV2Huge(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 27, 3]
    out_channels = [352, 704, 1408, 2816]
    return ConvNeXtV2(**locals(), model_name="convnext_v2_huge", **kwargs)



import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, initializers, image_data_format
from keras_cv_attention_models.attention_layers import (
    add_with_layer_scale_and_drop_block,
    ChannelAffine,
    CompatibleExtractPatches,
    conv2d_no_bias,
    drop_block,
    layer_norm,
    mlp_block,
    output_block,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

PRETRAINED_DICT = {
    "nat_base": {"imagenet": "d221eaec4af71dd3522625333aa73d9e"},
    "nat_mini": {"imagenet": "921de737ccebe4ab210dcd79ec0aed5f"},
    "nat_small": {"imagenet": "5716d0d54f3abd582e586fbdea04b3db"},
    "nat_tiny": {"imagenet": "91ed860ca950de181f433c189135070b"},
}


@backend.register_keras_serializable(package="kecam/nat")
class MultiHeadRelativePositionalKernelBias(layers.Layer):
    def __init__(self, input_height=-1, is_heads_first=False, **kwargs):
        super().__init__(**kwargs)
        self.input_height, self.is_heads_first = input_height, is_heads_first

    def build(self, input_shape):
        # input (is_heads_first=False): `[batch, height * width, num_heads, ..., size * size]`
        # input (is_heads_first=True): `[batch, num_heads, height * width, ..., size * size]`
        blocks, num_heads = (input_shape[2], input_shape[1]) if self.is_heads_first else (input_shape[1], input_shape[2])
        size = int(np.sqrt(float(input_shape[-1])))
        height = self.input_height if self.input_height > 0 else int(np.sqrt(float(blocks)))
        width = blocks // height
        pos_size = 2 * size - 1
        initializer = initializers.truncated_normal(stddev=0.02)
        self.pos_bias = self.add_weight(name="positional_embedding", shape=(num_heads, pos_size * pos_size), initializer=initializer, trainable=True)

        idx_hh, idx_ww = np.arange(0, size), np.arange(0, size)
        coords = np.reshape(np.expand_dims(idx_hh, -1) * pos_size + idx_ww, [-1]).astype("int64")
        bias_hh = np.concatenate([idx_hh[: size // 2], np.repeat(idx_hh[size // 2], height - size + 1), idx_hh[size // 2 + 1 :]], axis=-1)
        bias_ww = np.concatenate([idx_ww[: size // 2], np.repeat(idx_ww[size // 2], width - size + 1), idx_ww[size // 2 + 1 :]], axis=-1)
        bias_hw = np.expand_dims(bias_hh, -1) * pos_size + bias_ww
        bias_coords = np.expand_dims(bias_hw, -1) + coords
        bias_coords = np.reshape(bias_coords, [-1, size**2])[::-1]  # torch.flip(bias_coords, [0])

        bias_coords_shape = [bias_coords.shape[0]] + [1] * (len(input_shape) - 4) + [bias_coords.shape[1]]
        bias_coords = np.reshape(bias_coords, bias_coords_shape)  # [height * width, 1 * n, size * size]
        if hasattr(self, "register_buffer"):  # PyTorch
            self.register_buffer("bias_coords", functional.convert_to_tensor(bias_coords, dtype="int64"), persistent=False)
        else:
            self.bias_coords = functional.convert_to_tensor(bias_coords, dtype="int64")

        if not self.is_heads_first:
            self.transpose_perm = [1, 0] + list(range(2, len(input_shape) - 1))  # transpose [num_heads, height * width] -> [height * width, num_heads]

    def call(self, inputs):
        if self.is_heads_first:
            return inputs + functional.gather(self.pos_bias, self.bias_coords, axis=-1)
        else:
            return inputs + functional.transpose(functional.gather(self.pos_bias, self.bias_coords, axis=-1), self.transpose_perm)

    def get_config(self):
        base_config = super().get_config()
        base_config.update({"input_height": self.input_height, "is_heads_first": self.is_heads_first})
        return base_config


def neighborhood_attention(
    inputs, kernel_size=7, num_heads=4, key_dim=0, out_weight=True, qkv_bias=True, out_bias=True, attn_dropout=0, output_dropout=0, name=None
):
    _, hh, ww, cc = inputs.shape
    key_dim = key_dim if key_dim > 0 else cc // num_heads
    qk_scale = 1.0 / (float(key_dim) ** 0.5)
    out_shape = cc
    qkv_out = num_heads * key_dim

    should_pad_hh, should_pad_ww = max(0, kernel_size - hh), max(0, kernel_size - ww)
    if should_pad_hh or should_pad_ww:
        inputs = functional.pad(inputs, [[0, 0], [0, should_pad_hh], [0, should_pad_ww], [0, 0]])
        _, hh, ww, cc = inputs.shape

    qkv = layers.Dense(qkv_out * 3, use_bias=qkv_bias, name=name and name + "qkv")(inputs)
    query, key_value = functional.split(qkv, [qkv_out, qkv_out * 2], axis=-1)  # Matching weights from PyTorch
    query = functional.expand_dims(functional.reshape(query, [-1, hh * ww, num_heads, key_dim]), -2)  # [batch, hh * ww, num_heads, 1, key_dim]

    # key_value: [batch, height // kernel_size, width // kernel_size, kernel_size, kernel_size, key + value]
    kv = CompatibleExtractPatches(sizes=kernel_size, strides=1, padding="VALID", compressed=False)(key_value)
    padded = (kernel_size - 1) // 2
    # torch.pad 'replicate'
    kv = functional.concat([functional.repeat(kv[:, :1], padded, axis=1), kv, functional.repeat(kv[:, -1:], padded, axis=1)], axis=1)
    key_value = functional.concat([functional.repeat(kv[:, :, :1], padded, axis=2), kv, functional.repeat(kv[:, :, -1:], padded, axis=2)], axis=2)

    key_value = functional.reshape(key_value, [-1, kernel_size * kernel_size, key_value.shape[-1]])
    key, value = functional.split(key_value, 2, axis=-1)  # [batch * block_height * block_width, K * K, key_dim]
    key = functional.transpose(functional.reshape(key, [-1, key.shape[1], num_heads, key_dim]), [0, 2, 3, 1])  # [batch * hh*ww, num_heads, key_dim, K * K]
    key = functional.reshape(key, [-1, hh * ww, num_heads, key_dim, kernel_size * kernel_size])  # [batch, hh*ww, num_heads, key_dim, K * K]
    value = functional.transpose(functional.reshape(value, [-1, value.shape[1], num_heads, key_dim]), [0, 2, 1, 3])
    value = functional.reshape(value, [-1, hh * ww, num_heads, kernel_size * kernel_size, key_dim])  # [batch, hh*ww, num_heads, K * K, key_dim]
    # print(f">>>> {query.shape = }, {key.shape = }, {value.shape = }")

    # [batch, hh * ww, num_heads, 1, kernel_size * kernel_size]
    # attention_scores = layers.Lambda(lambda xx: tf.matmul(xx[0], xx[1]))([query, key]) * qk_scale
    attention_scores = (query @ key) * qk_scale
    attention_scores = MultiHeadRelativePositionalKernelBias(input_height=hh, name=name and name + "pos")(attention_scores)
    attention_scores = layers.Softmax(axis=-1, name=name and name + "attention_scores")(attention_scores)
    attention_scores = layers.Dropout(attn_dropout, name=name and name + "attn_drop")(attention_scores) if attn_dropout > 0 else attention_scores

    # attention_output = [batch, block_height * block_width, num_heads, 1, key_dim]
    # attention_output = layers.Lambda(lambda xx: tf.matmul(xx[0], xx[1]))([attention_scores, value])
    attention_output = attention_scores @ value
    attention_output = functional.reshape(attention_output, [-1, hh, ww, num_heads * key_dim])
    # print(f">>>> {attention_output.shape = }, {attention_scores.shape = }")

    if should_pad_hh or should_pad_ww:
        attention_output = attention_output[:, : hh - should_pad_hh, : ww - should_pad_ww, :]

    if out_weight:
        # [batch, hh, ww, num_heads * key_dim] * [num_heads * key_dim, out] --> [batch, hh, ww, out]
        attention_output = layers.Dense(out_shape, use_bias=out_bias, name=name and name + "output")(attention_output)
    attention_output = layers.Dropout(output_dropout, name=name and name + "out_drop")(attention_output) if output_dropout > 0 else attention_output
    return attention_output


def nat_block(inputs, attn_kernel_size=7, num_heads=4, mlp_ratio=4, mlp_drop_rate=0, attn_drop_rate=0, drop_rate=0, layer_scale=-1, name=None):
    input_channel = inputs.shape[-1]

    attn = layer_norm(inputs, name=name + "attn_")
    attn = neighborhood_attention(attn, attn_kernel_size, num_heads, attn_dropout=attn_drop_rate, name=name + "attn_")
    attn_out = add_with_layer_scale_and_drop_block(inputs, attn, layer_scale=layer_scale, drop_rate=drop_rate, name=name + "1_")

    mlp = layer_norm(attn_out, name=name + "mlp_")
    mlp = mlp_block(mlp, int(input_channel * mlp_ratio), drop_rate=mlp_drop_rate, activation="gelu", name=name + "mlp_")
    return add_with_layer_scale_and_drop_block(attn_out, mlp, layer_scale=layer_scale, drop_rate=drop_rate, name=name + "2_")


def NAT(
    num_blocks=[3, 4, 6, 5],
    out_channels=[64, 128, 256, 512],
    num_heads=[2, 4, 8, 16],
    stem_width=-1,
    attn_kernel_size=7,
    mlp_ratio=3,
    layer_scale=-1,
    input_shape=(224, 224, 3),
    num_classes=1000,
    drop_connect_rate=0,
    classifier_activation="softmax",
    dropout=0,
    pretrained=None,
    model_name="nat",
    kwargs=None,
):
    """ConvTokenizer stem"""
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimention is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)
    stem_width = stem_width if stem_width > 0 else out_channels[0]
    nn = conv2d_no_bias(inputs, stem_width // 2, kernel_size=3, strides=2, use_bias=True, padding="SAME", name="stem_1_")
    nn = conv2d_no_bias(nn, stem_width, kernel_size=3, strides=2, use_bias=True, padding="SAME", name="stem_2_")
    nn = layer_norm(nn, name="stem_")

    """ stages """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, (num_block, out_channel, num_head) in enumerate(zip(num_blocks, out_channels, num_heads)):
        stack_name = "stack{}_".format(stack_id + 1)
        if stack_id > 0:
            ds_name = stack_name + "downsample_"
            nn = conv2d_no_bias(nn, out_channel, kernel_size=3, strides=2, padding="SAME", name=ds_name)
            nn = layer_norm(nn, name=ds_name)
        for block_id in range(num_block):
            block_name = stack_name + "block{}_".format(block_id + 1)
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            nn = nat_block(nn, attn_kernel_size, num_head, mlp_ratio, drop_rate=block_drop_rate, layer_scale=layer_scale, name=block_name)
            global_block_id += 1
    nn = layer_norm(nn, name="pre_output_")

    nn = output_block(nn, num_classes=num_classes, drop_rate=dropout, classifier_activation=classifier_activation)
    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, PRETRAINED_DICT, "nat", pretrained)
    return model


def NAT_Mini(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 6, 5]
    return NAT(**locals(), model_name="nat_mini", **kwargs)


def NAT_Tiny(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 18, 5]
    return NAT(**locals(), model_name="nat_tiny", **kwargs)


def NAT_Small(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 18, 5]
    num_heads = [3, 6, 12, 24]
    out_channels = [96, 192, 384, 768]
    mlp_ratio = kwargs.pop("mlp_ratio", 2)
    layer_scale = kwargs.pop("layer_scale", 1e-5)
    return NAT(**locals(), model_name="nat_small", **kwargs)


def NAT_Base(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 18, 5]
    num_heads = [4, 8, 16, 32]
    out_channels = [128, 256, 512, 1024]
    mlp_ratio = kwargs.pop("mlp_ratio", 2)
    layer_scale = kwargs.pop("layer_scale", 1e-5)
    return NAT(**locals(), model_name="nat_base", **kwargs)



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format
from keras_cv_attention_models.attention_layers import (
    add_with_layer_scale_and_drop_block,
    ChannelAffine,
    MultiHeadPositionalEmbedding,
    batchnorm_with_activation,
    conv2d_no_bias,
    drop_block,
    layer_norm,
    mlp_block,
    mhsa_with_multi_head_position,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

PRETRAINED_DICT = {
    "efficientformer_l1": {"imagenet": {224: "7698d40d502ccc548a7e2890fb33db34"}},
    "efficientformer_l3": {"imagenet": {224: "ee3d11742d233bc2ec36648440cb5a0b"}},
    "efficientformer_l7": {"imagenet": {224: "66c26fc1e0bd39bbf6886d570956d178"}},
}


def attn_block(inputs, num_heads=8, key_dim=32, attn_ratio=4, mlp_ratio=4, layer_scale=0, drop_rate=0, activation="gelu", name=""):
    input_channel = inputs.shape[-1]  # also using -1 for channels_first

    nn = layer_norm(inputs, axis=-1, name=name + "attn_")
    nn = mhsa_with_multi_head_position(nn, num_heads, key_dim=key_dim, attn_ratio=attn_ratio, qkv_bias=True, out_bias=True, name=name)
    attn_out = add_with_layer_scale_and_drop_block(inputs, nn, layer_scale=layer_scale, drop_rate=drop_rate, axis=-1, name=name + "attn_")

    nn = layer_norm(attn_out, axis=-1, name=name + "mlp_")
    nn = mlp_block(nn, input_channel * mlp_ratio, activation=activation, name=name)
    return add_with_layer_scale_and_drop_block(attn_out, nn, layer_scale=layer_scale, drop_rate=drop_rate, axis=-1, name=name + "mlp_")


def conv_block(inputs, mlp_ratio=4, layer_scale=0, drop_rate=0, activation="gelu", name=""):
    input_channel = inputs.shape[-1 if image_data_format() == "channels_last" else 1]

    nn = layers.AvgPool2D(pool_size=3, strides=1, padding="SAME")(inputs)  # count_include_pad=False [ ??? ]
    nn = nn - inputs
    attn_out = add_with_layer_scale_and_drop_block(inputs, nn, layer_scale=layer_scale, drop_rate=drop_rate, name=name + "attn_")

    nn = conv2d_no_bias(attn_out, input_channel * mlp_ratio, 1, strides=1, use_bias=True, name=name + "mlp_1_")
    nn = batchnorm_with_activation(nn, activation=activation, name=name + "mlp_1_")
    nn = conv2d_no_bias(nn, input_channel, 1, strides=1, use_bias=True, name=name + "mlp_2_")
    nn = batchnorm_with_activation(nn, activation=None, name=name + "mlp_2_")
    return add_with_layer_scale_and_drop_block(attn_out, nn, layer_scale=layer_scale, drop_rate=drop_rate, name=name + "mlp_")


def EfficientFormer(
    num_blocks=[3, 2, 6, 4],
    out_channels=[48, 96, 224, 448],
    mlp_ratios=4,
    num_attn_blocks_each_stack=[0, 0, 0, 1],
    stem_width=-1,
    stem_activation="relu",
    layer_scale=1e-5,
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu",
    drop_connect_rate=0,
    classifier_activation=None,
    use_distillation=True,
    dropout=0,
    pretrained=None,
    model_name="efficientformer",
    kwargs=None,
):
    # Regard input_shape as force using original shape if first element is None or -1,
    # else assume channel dimention is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)
    stem_width = stem_width if stem_width > 0 else out_channels[0]
    stem_activation = stem_activation if stem_activation is not None else activation
    nn = conv2d_no_bias(inputs, stem_width // 2, 3, strides=2, use_bias=True, padding="same", name="stem_1_")
    nn = batchnorm_with_activation(nn, activation=stem_activation, name="stem_1_")
    nn = conv2d_no_bias(nn, stem_width, 3, strides=2, use_bias=True, padding="same", name="stem_2_")
    nn = batchnorm_with_activation(nn, activation=stem_activation, name="stem_2_")

    """ stages """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, (num_block, out_channel) in enumerate(zip(num_blocks, out_channels)):
        stack_name = "stack{}_".format(stack_id + 1)
        if stack_id > 0:
            ds_name = stack_name + "downsample_"
            nn = conv2d_no_bias(nn, out_channel, kernel_size=3, strides=2, use_bias=True, padding="SAME", name=ds_name)
            nn = batchnorm_with_activation(nn, activation=None, name=ds_name)

        cur_mlp_ratios = mlp_ratios[stack_id] if isinstance(mlp_ratios, (list, tuple)) else mlp_ratios
        cur_num_attn_blocks = num_attn_blocks_each_stack[stack_id] if isinstance(num_attn_blocks_each_stack, (list, tuple)) else num_attn_blocks_each_stack
        attn_block_start_id = num_block - cur_num_attn_blocks
        for block_id in range(num_block):
            block_name = stack_name + "block{}_".format(block_id + 1)
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            mlp_ratio = cur_mlp_ratios[block_id] if isinstance(cur_mlp_ratios, (list, tuple)) else cur_mlp_ratios
            if block_id >= attn_block_start_id:
                nn = layers.Permute([2, 3, 1])(nn) if block_id == attn_block_start_id and image_data_format() == "channels_first" else nn
                nn = attn_block(nn, mlp_ratio=mlp_ratio, layer_scale=layer_scale, drop_rate=block_drop_rate, activation=activation, name=block_name)
                nn = layers.Permute([3, 1, 2])(nn) if block_id == num_block - 1 and image_data_format() == "channels_first" else nn
            else:
                nn = conv_block(nn, mlp_ratio=mlp_ratio, layer_scale=layer_scale, drop_rate=block_drop_rate, activation=activation, name=block_name)

            global_block_id += 1

    """ output """
    if num_classes > 0:
        nn = layer_norm(nn, name="pre_output_")
        nn = layers.GlobalAveragePooling2D()(nn)  # tf.reduce_mean(nn, axis=1)
        if dropout > 0 and dropout < 1:
            nn = layers.Dropout(dropout)(nn)
        out = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="head")(nn)

        if use_distillation:
            distill = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="distill_head")(nn)
            out = [out, distill]
    else:
        out = nn

    model = models.Model(inputs, out, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, PRETRAINED_DICT, "efficientformer", pretrained, MultiHeadPositionalEmbedding)
    return model


def EfficientFormerL1(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", use_distillation=True, pretrained="imagenet", **kwargs):
    return EfficientFormer(**locals(), model_name="efficientformer_l1", **kwargs)


def EfficientFormerL3(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", use_distillation=True, pretrained="imagenet", **kwargs):
    num_blocks = [4, 4, 12, 6]
    out_channels = [64, 128, 320, 512]
    num_attn_blocks_each_stack = [0, 0, 0, 4]
    return EfficientFormer(**locals(), model_name="efficientformer_l3", **kwargs)


def EfficientFormerL7(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", use_distillation=True, pretrained="imagenet", **kwargs):
    num_blocks = [6, 6, 18, 8]
    out_channels = [96, 192, 384, 768]
    num_attn_blocks_each_stack = [0, 0, 0, 8]
    return EfficientFormer(**locals(), model_name="efficientformer_l7", **kwargs)
